#! /usr/bin/env python3

# MagPy
# Author:   TG Tan      www.pestobservatory.com
#
# V0.0      13/04/15    Draft version
# V1.00     03/05/15    First working version, only ensemble photometry
# V1.01     04/05/15    Full working version doing only ensemble photometry
# V1.1      10/05/15    Revise to make the 'calculate' function take the ensemble string as argument
#                       This enables production of the std photometry calcs, and optimisation of the ensemble
# V1.11     11/05/15    Revise to write a bash script that calls makelc to plot results
# V1.12     27/05/15    Revise to include a button that fills in the current directory as path
# V1.2      18/07/16    Converted to python3
# V1.21     1/08/16     Bug fix; convert the float 'num_stars' to int
# V1.22     7/08/16     Bug fixes to handle; blank ensemble list, previous directory was deleted
# V1.3      7/08/16     Plots light curves without having to close the script
# V1.31     8/1/17      Bug fix to correct entry field for Group and order of items in AAVSO file
# V1.4      2/11/17     Revise to only work in the current directory, removing the Browse and Cur Dir buttons
# V1.41     6/11/17     Bug fix for crash when a star has only 1 valid observation
# V1.42     21/11/17    Check and populate filter, target name.  Hardcode start of input filenames as 'Muni'
# V1.43     21/11/17    Add vertical lines on LC plot for VAR to mark ingress and egress
# V1.44     2/1/18      Accept BJD, and modify plotting to show the date type used (previously only HJD)
# V1.5      12/1/18     Mod GUI to allow choice of whether to use detrended ensemble stars,
#                       note on ENSEMBLE_STATS file whether de-trended ensembles were used.
# V1.6      4/4/18      Add production of a file that combines photometry for the target and de-trend data
# V2.0      19/4/18     Read customisation data from the ~/bin/pest_pipeline_config.txt file
# V3.0      28/4/18     Do LC plotting within python, rather than Gnuplot
#                       Place working files into a single subdirectory, also all star data files
#                       Fix COMP and CHK star nos. Use star ids in ensemble and file names
# V3.1      2/5/18      Add twin axis to lightcurves with delta-mag, added error handling in reading config file
# V3.2      10/7/18     Fix for overlapping x axis labels
# V3.3      21/10/18    Read photometry rad, fwhm and exp time from photcorr_info.txt. Improve report template.
# V3.4      24/10/18    Add an output file to help fill out TESS ExoFOP Time Series entry, TESS_time_series_data.txt
# V3.5      29/11/18    Minor change to plotting routine to reduce calls to float()
# V3.6      4/12/18     Bug fix for plotting for very short obs durations - when tick_spacing might == 0
# V3.7      5/12/18     Make paths platform agnostic,ie get rid of linux /. Eliminate unnecessary files.
# V3.8      8/1/19      Add facility to plot data points binned/blocked together. Remove unnecessary imports
# V3.9      11/1/19     Bug fix for case where the first datapoint has invalid COMP mag
# V4.0      20/1/19     Improve appearance of GUI using themes
# V4.1      28/1/19     Minor change - default block size of 270s, rather than 0
# V4.11     30/1/19     Minor change - fix observatory label on plots spilling out of image area
# V4.2      26/4/19     Re-enable use of HJD date type, multiple input files, and AAVSO reports
# V4.3      28/4/19     Enable reading of COMP mag and ingress/egress info if that was input to ebcheck
# V4.4      9/5/19      Add max_dmag to TESS_time_series.txt
# V4.5      24/5/19     Remove TESS_time_series.txt.  Now written by PESTphot.
#                       photcorr_info.txt --> PESTphot_info.txt, mag_times.txt --> PESTphot_inputs.txt
#                       find and copy over the select_labels.txt and PESTphot_info.txt files if not present
# V4.6      30/5/19     For multi session data, plot small points without errorbars
# V5.0      31/5/19     Major rev. for speed. Reduce file writes, use data structures instead, re-order code
# V5.1      2/6/19      Debug code order overwriting correct parameters
# V5.2      4/6/19      Drop ensemble stars that have too many invalid points. Check for existing INFO.txt
#                       Debug not using de-trended data.  Debug where PESTphot_info.txt not present.
# V5.3      12/6/19     Remove the files INSTR_MAG_ERR_AVERAGES.txt and WT_AVG_MAG.txt
# V5.4      18/6/19     Added tickbox to remove hyphen in target name for AAVSO report
#
# Script Function:
#  Reads all text files (Muniwin format) matching a search string in the input directory.
#  These files contain magnitudes and std errors of selected stars.  The script parses these files and produces
#  a number of files, each corresponding to one star,
#  containing times, magnitudes and errors based lines to the COMP star, with all observation sets
#  concatenated together.
#
#  ENSEMBLE files are produced that have magnitudes and errors adjusted based on an ensemble of stars.
#  The stars in the ensemble are specified.
#
#  The sequence for the ensemble part of the processing is as follows#
#  - concatenate all Muniwin format files into the list instr_concat_lines
#  - calculate the weighted average instr mag of all ensemble stars and subtract that from the instr mag of each star
#  - record in the list reduced_mag_lines
#  - shift these instr mags to the standard scale such that the mean mag of the COMP star is equal to the catalog mag,
#  and parse to produce one data file for each star
#
#  
#  Other files produced:
#  1.  an INFO file that contains the parameters used and which can replace the MagPy_param.txt file for a repeat run.
#  2.  an ENSEMBLE_STATS file with info on the Mean and Std Dev of each star post ensemble processing
#  3.  an AAVSO_OUT file that presents data in the AAVSO format
#  4.  mag+detrend-data.txt that combines mag, err for the target with corresponding detrend data
#
#
# Assumptions:
#
# 1.  Input file format (.txt) is JDHEL MAG0 ERR0 MAG1 ERR1 MAG2 ERR2 MAG3 ERR3 MAG4 ERR4 ... <AIRMASS>
# 2.  First 2 lines in each input file are ignored.
# 3.  Airmass is required
# 4.  Current working directory contains input files, output files written there, as well as subdir /magpy_files
# 5.  Input filenames must start with 'Muni', or 'det_Muni' for detrended ensemble stars
# 6.  *.detrend-data.txt file as produced by PESTphot exists
# 7.  The ~/bin/pest_pipeline_config.txt file exists
#
#########################################################################################
#
# Documentation of data structures:
#
# instr_concat_lines  : a list that contains all data lines from all Muni files
# reduced_mag_lines : a list that corresponds to each line in instr_concat_lines, but with reduced mags
# time_list_ens, mag_list_ens, err_list_ens : lists that hold data for an ensemble run
# time_list_std, mag_list_std, err_list_std : lists that hold data for an ensemble run
#
#########################################################################################


from tkinter import *
import tkinter.ttk, tkinter.filedialog, math, os, glob, tkinter.messagebox, subprocess, re
from tkinter import ttk
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from PIL import Image
from shutil import copyfile


script_name = "MagPy by TG Tan: "
ver_no = "V5.4"

##########################################################################################
# Function to return a variable value given the corresponding keyword in
# the pest_pipeline_config.txt file
##########################################################################################

def get_conf_var(keyword, conf_list, default):  # keyword is string, conf_list is the list read from the config file
    expr = keyword + '\s*=\s*(.*)'      # will match 'keyword= variable' with or without spaces next to the =
    var_found = False
    for line in conf_list:
        try:
            var = re.search(expr, line).group(1)
            var_found = True
            break
        except:
            pass

    if not var_found:
        print(keyword + ' not found in pest_pipeline_config.txt.\nSetting to: ' + default)
        var = default
    return var

# end get_conf_var


##################################################################################################
#
# Function that returns the error due to scintillation for a given airmass
#
##################################################################################################

def Scint_Err(airmass, exp_time):
    global scope_dia
    err = 0.09 * (airmass**1.75)/((float(scope_dia)/10.0)**0.66 * math.sqrt(2.0*exp_time))  # decimal pts to ensure float type
    return err

##################################################################################################
#
# End Scint_Err
#
##################################################################################################




####################################################################################################
# Function to plot a light curve
####################################################################################################
def plot_lc(bjd, mag, err, meanmag, std, title, filename, ingress, egress):
    # title, filename, ingress, egress are strings, meanmag, std: float, the rest arrays

    bjd_start = np.min(bjd)
    fingress = float(ingress)
    fegress = float(egress)

    fig, ax1 = plt.subplots(figsize=(12,6))
    plt.rc('grid', linestyle='dotted', color='gray')

    if multi:
        ax1.plot(bjd - 2450000, mag - meanmag, 'b.', alpha=0.2)
    else:
        ax1.errorbar(bjd - 2450000, mag-meanmag, yerr=err, fmt='bo', lw=1, capsize=3, alpha=0.5)

    ax1.get_xaxis().get_major_formatter().set_useOffset(False) # turn off scientific notation
    ax1.get_yaxis().get_major_formatter().set_useOffset(False) # turn off scientific notation

    ax1.invert_yaxis()
    ax1.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.3f'))

    # allow some margin around the data
    ax1.margins(0.05)

    start, end = ax1.get_xlim()  # get the start and end limits of the x-axis
    low, high = ax1.get_ylim()

    ax1.set_ylim(low, high)  # for some reason, get does not return the correct values. To align left and right
    # y axes this had to be done.

    tick_spacing = math.floor((end-start)*10.0)/100  # makes sure there are about 10 ticks on the x-axis
    if tick_spacing > 0:    # unless there is a very short observation duration
        ax1.xaxis.set_major_locator(ticker.MultipleLocator(tick_spacing))

    # plot a horizontal at the mean of the data
    plt.axhline(y=0.0, color='gray', linestyle='--', lw=2)

    # find where along the y-axis to put the ingress and egress labels
    y_pos = low

    # plot vertical lines to mark ingress and egress
    # note the adjustments to the text coords are due to us putting a margin around the data
    if ingress != '0':
        plt.axvline(x=fingress, color='green', linestyle='--', lw=1, ymin=0.1)
        ax1.annotate('Pred\nIngress', xy=(fingress, y_pos), xytext=(fingress, y_pos),
                     horizontalalignment='center', verticalalignment='bottom', color='green')

    if egress != '0':
        plt.axvline(x=fegress, color='red', linestyle='--', lw=1, ymin=0.1)
        ax1.annotate('Pred\nEgress', xy=(fegress, y_pos), xytext=(fegress, y_pos),
                     horizontalalignment='center', verticalalignment='bottom', color='red')


    plt.title(title)
    if date_type == 'HJD':
        ax1.set_xlabel('HJD (UTC) - 2450000')
    else:
        ax1.set_xlabel('BJD (TDB) - 2450000')

    ax1.set_ylabel('delta-mag')
    ax1.figure.autofmt_xdate()  # to avoid overlapping x labels


    # make the Mag axis
    ax2 = ax1.twinx()
    ax2.set_ylim(low+meanmag, high+meanmag)
    ax2.set_ylabel('Mag')
    ax1.grid(True)
    ax2.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.3f'))



    plt.text(0.98, 0.95, 'Mean = ' + str(round(meanmag,3)), \
             horizontalalignment='right', transform=ax1.transAxes)
    plt.text(0.98, 0.9, 'SD = ' + str(round(std,4)), \
             horizontalalignment='right', transform=ax1.transAxes)

    plt.text(0.01, 0.95, 'Target: ' + obj_id + ' in ' + filterband, \
             horizontalalignment='left', transform=ax1.transAxes)

    plt.text(0.01, 0.9, 'Obs Date: ' + obs_date, \
             horizontalalignment='left', transform=ax1.transAxes)
    plt.text(0.0, -0.15, observatory + ' Observatory', color='gray', \
             horizontalalignment='left', transform=ax1.transAxes)

    ##############  calculate binned data points if exposure time is less than 61s ############

    if block_size != '0':
        max_bin_s = float(block_size)
        max_bin_time = max_bin_s/(24.*60.*60.)  # max time over which to bin, in fraction of a day

        list_bjd = []
        list_mag = []  # these lists will hold the binned BJD and mag

        sum_bjd = 0.0
        sum_mag = 0.0
        count = 0.0
        bin_start = bjd_start-2450000  # the start time

        for i_bjd, i_mag in zip(bjd-2450000, mag-meanmag):
            sum_bjd += i_bjd
            sum_mag += i_mag
            count += 1.0

            if i_bjd - bin_start > max_bin_time:
                list_bjd.append(sum_bjd/count)
                list_mag.append(sum_mag/count)
                sum_bjd = 0.0
                sum_mag = 0.0
                count = 0.0
                bin_start = i_bjd

        binned_bjd = np.array(list_bjd)
        binned_mag = np.array(list_mag)
        binned_sd = np.nanstd(binned_mag, ddof=1)

        ax1.plot(binned_bjd, binned_mag, 'rs')

        plt.text(0.01, 0.07, 'Max block size: ' + str(round(max_bin_s)) + 's',
                 color='red', horizontalalignment='left', transform=ax1.transAxes)
        plt.text(0.01, 0.02, 'SD = ' + str(round(binned_sd,4)),
                 color='red', horizontalalignment='left', transform=ax1.transAxes)




    plt.tight_layout()
    plt.savefig(filename,dpi=72)
    plt.close("all")        # clears memory

# End plot_lc()
###################################################################################################



##################################################################################################
#
# Function to convert a comma delimited string of star-id to star nos
#
##################################################################################################
def sid2sno(sid_str):
    sno_list = []
    sid_list = sid_str.split(',')
    for i in sid_list:
        try:
            s = np.where(star_ids==int(i))[0][0] + 1
            sno_list.append(str(s))
        except:
            print('Star id does not exist. Will be dropped: ' + i)
            pass

    sno_str = ','.join(sno_list)
    return sno_str

##################################################################################################

##################################################################################################
#
# Function to convert a comma delimited string of star nos to star-id
#
##################################################################################################
def sno2sid(sno_str):
    sid_list = []
    sno_list = sno_str.split(',')
    for s in sno_list:
        try:
            i = star_ids[int(s) - 1]
            sid_list.append(str(i))
        except:
            print('Warning: star no. out of bounds - ignoring.\n')
            pass

    sid_str = ','.join(sid_list)
    return sid_str

##################################################################################################

################################################################################################
#
# finish by combining the target mags with detrend data
# (Note that this assumes that all detrend parameters are used)
# also make lightcurves and write out the report file
#
###############################################################################################

def finish():
    global time_list_ens, mag_list_ens, err_list_ens
    global time_list_std, mag_list_std, err_list_std


    det_file = glob.glob('*_detrend-data.txt')    # the file produced by PhotCorr
    if len(det_file) == 1 and not multi:

        # load the file into a dataframe, skipping first 2 rows, using any whitespace as separator
        # the engine='python' is needed because the c engine cannot parse the regex "\s+"
        # if for some reason more than 1 file is found, only the first is used


        # dtype=object below forces all fields to be read as string, to make sure the merge happens right
        det_frame = pd.read_csv(det_file[0], sep="\s+", skiprows=2, header=None, engine='python',dtype=object)

        # create a zipped list of tuples in order to make the dataframe
        zipdata = list(zip(time_list_ens[1], mag_list_ens[1], err_list_ens[1]))
        target_frame = pd.DataFrame(zipdata)

        if date_type == 'HJD':
            det_frame.columns = ['frame_no', 'HJD', 'dist_center', 'fwhm', 'airmass', 'sky']
            det_frame = det_frame.drop('frame_no', 1)
            target_frame.columns = ['HJD', 'mag', 'mag_err']
            merged_frame = pd.merge(target_frame, det_frame, on='HJD')
        else:
            det_frame.columns = ['frame_no', 'BJD_TDB', 'dist_center', 'fwhm', 'airmass', 'sky']
            det_frame = det_frame.drop('frame_no', 1)
            target_frame.columns = ['BJD_TDB', 'mag', 'mag_err']
            merged_frame = pd.merge(target_frame, det_frame, on='BJD_TDB')

        # now write the merged frame to file
        data_format = [lambda x: x, lambda x: "%.4f" % x, lambda x: "%.4f" % x, lambda x: x,
                       lambda x: x, lambda x: x, lambda x: x]
        f_mag_detrend = open(std_name+'_mag+detrend-data.txt', 'w')
        f_mag_detrend.write('#' + merged_frame.to_string(header=True, index=None, formatters=data_format).lstrip())
        # this is a bit of formatting. Put # at the start of the line and remove leading whitespaces in the rest

        f_mag_detrend.close()
    else:
        print('\nWARNING:  *detrend-data.txt not found, or multi input run. Mag+detrend file not produced.\n')

    ##############################################################################################
    # Plot light curves
    ##############################################################################################

    plot_list = []  # list of all plot filenames, to enable combination into a multiple plot

    for i in range(len(star_ids)):
        sid = star_ids[i]
        sno = i+1
        label = star_labels[i]

        plot_filename = std_name + '_LC' + str(sno) + '-StarID-' + str(sid) + '.png'

        float_time = [float(x) for x in time_list_ens[sno]]
        zipdata = list(zip(float_time, mag_list_ens[sno], err_list_ens[sno]))
        lc_frame = pd.DataFrame(zipdata)

        lc_frame.columns = ['bjd', 'mag', 'err']

        lc_frame_sort = lc_frame.sort_values('bjd')  # needed to enable block plotting for 'multi'
        bjd = lc_frame_sort['bjd'].values
        mag = lc_frame_sort['mag'].values
        err = lc_frame_sort['err'].values

        meanmag = np.nanmean(mag)
        std = np.nanstd(mag, ddof=1)

        if label == 'var':
            title = 'Target(VAR) Star ID: ' + str(sid)
        elif label == 'comp':
            title = 'Comparison(COMP) Star ID: ' + str(sid)
        else:
            title = 'Field star Star ID: ' + str(sid)

        plot_list.append(plot_filename)
        plot_lc(bjd,mag, err, meanmag, std, title, plot_filename, ingress, egress)

        # write star data files
        data_format = [lambda x: "%.7f" % x, lambda x: "%.4f" % x, lambda x: "%.4f" % x]
        f_data = open(os.path.join('magpy_files', 'Star-' + str(sid) + '_' + obs_date_strip + '_' +
                                   observatory + '_' + filterband + '_data.txt'), 'w')
        f_data.write(lc_frame_sort.to_string(header=False, index=None, formatters=data_format))
        f_data.close()

        ############################
        # this section for de-bugging - produces data files for the STD runs
        '''
        float_time_std = [float(x) for x in time_list_std[sno]]
        zipdata = list(zip(float_time_std, mag_list_std[sno], err_list_std[sno]))
        std_frame = pd.DataFrame(zipdata)

        std_frame.columns = ['bjd', 'mag', 'err']
        std_frame_sort = std_frame.sort_values('bjd')  # needed to enable block plotting for 'multi'

        f_data = open(os.path.join('magpy_files', 'std_Star-' + str(sid) + '_' + obs_date_strip + '_' +
                                   observatory + '_' + filterband + '_data.txt'), 'w')
        f_data.write(std_frame_sort.to_string(header=False, index=None, formatters=data_format))
        f_data.close()
        '''





    ############################################################################################
    # combine several plots into 1 image file

    images = [Image.open(x) for x in [plot_list[1], plot_list[2], plot_list[3], plot_list[4]]]
    widths, heights = zip(*(i.size for i in images))
    mode = images[0].mode

    total_height = sum(heights)
    max_width = max(widths)

    new_im = Image.new(mode, (max_width, total_height))

    y_offset = 0
    for im in images:
      new_im.paste(im, (0, y_offset))
      y_offset += im.size[1]

    new_im.save(std_name+'_sample-lightcurves.png')

    ###########################################################################################
    # Write the template Comments file
    ###########################################################################################

    if multi:
        obs_date_report = ' ***Multiple***'
        obs_dur_report = '***Multiple***'
    else:
        obs_date_report = obs_date
        obs_dur_report = obs_dur


    f_comments = open(std_name + '_Comments.txt', 'w')
    f_comments.write('Photometric follow-up: '+obj_id+' on UT'+obs_date_report+' from '+observatory+
                     ' in '+filterband+'\n\n')
    f_comments.write('Comments:\n\n\n')
    f_comments.write('Photometry aperture '+str(round(phot_rad_arcsec,1))+'", FWHM = '+fwhm+
                     '", Obs duration '+obs_dur_report+' mins.\n\n')
    f_comments.write('Target: '+obj_id+' (var=star#'+str(var_id)+');\n')
    f_comments.write('Date: '+obs_date_report+';\n')
    f_comments.write('Observatory: '+observatory+';\n')
    f_comments.write('Filter: '+filterband+';\n')
    f_comments.write('Exposure time: '+str(round(float(exp_time),1))+'s;\n')
    if date_type == 'HJD':
        f_comments.write('Data file format: HJD(UTC) Mag Err;\n')
    else:
        f_comments.write('Data file format: BJD(TDB) Mag Err;\n')
    f_comments.write('Ensemble star ids: '+sno2sid(ensemble_list)+'\n')
    f_comments.write('\n\n------------------------------------\n')
    f_comments.write(signature.strip('\'\"').replace('\\n', '\n')+'\n') # removes quotes, and convert \n to newline
    f_comments.close()

    #############################################################################################
    # copy the target data file to the main directory and give it the standard name

    copyfile(os.path.join('magpy_files', 'Star-' + str(var_id) + '_' + obs_date_strip + '_' +
                          observatory + '_' + filterband + '_data.txt'), std_name+'_data.txt')



# end finish()
#############################################################################################################

###################################################################################################
#
# The RUN button calls this part of the script, which modifies the ensemble list to get standard
# photometry, and then to optimise the ensemble list.
#
###################################################################################################

def controller():
    global ensemble_list_str, run_status_str, comp_star_no, num_stars, optimise
    global better_stars, detrend, in_filename_template, block_size, opt_done, mag_list_std

    # the minimum number of valid points for a star to be accepted in the ensemble
    # expressed as a fraction of the number of valid points for the target
    ens_star_min_frac = 0.95

    opt_done = False  # marks when optimisation is completed
    ensemble_sid_list = ensemble_list_str.get()

    # translate the star ids in the input to star nos
    ensemble_list = sid2sno(ensemble_sid_list)

    block_size = block_str.get()

    #comp_star_no = '2'  # star no 2 is always COMP

    if detrend.get():
        in_filename_template = 'det_Muni'  # use Muni file with detrended ensemble stars
    else:
        in_filename_template = 'Muni'

    if ensemble_list.strip() == '':  # this catches case if ensemble list is left blank. If so set it to COMP
        ensemble_list = comp_star_no

    if optimise.get():  # The Optimise checkbox is ticked
        ensemble_list_str.set(sno2sid(comp_star_no))
        calculate("STD")  # run to get standard photometry
        ensemble_list_str.set(sno2sid(ensemble_list))
        calculate("ENS")  # run once to set 'better_stars'

        # The optimisation method is to start with the 'better_stars' in the ensemble, then drop the one
        # that results in the best improvement in the optimisation criteria.
        # Repeat until removing an ensemble star results in a worse outcome.

        remain_ens = []  # will hold the list of remaining ensemble stars
        for ens in better_stars:  # Initialise
            if len(mag_list_std[ens])/len(mag_list_std[1]) > ens_star_min_frac: # use only the good stars
                remain_ens.append(str(ens))

        f_opt = open(os.path.join(in_path, 'magpy_files', 'OPTIMISATION.txt'), 'w')

        while len(remain_ens) > 1:
            improved = 0  # marks if dropping any star from the current ensemble leads to improvement
            curr_ens_list = ','.join(remain_ens)  # assemble the ensemble list to be used
            ensemble_list_str.set(sno2sid(curr_ens_list))
            base_result = calculate("ENS")  # this run uses the full set of remaining ensemble stars
            f_opt.write("Opt Criteria: " + str(round(base_result, 0)) + "\tEnsemble: "
                        + curr_ens_list + "\n")

            # next we remove one star at a time from the ensemble and test to see which gives the
            # most improvement in the optimisation criteria

            best_so_far = base_result
            for ens in remain_ens:
                ens_list = remain_ens[:]  # make a copy of all elements into ens_list
                ens_list.remove(ens)
                ens_list_joined = ','.join(ens_list)
                ensemble_list_str.set(sno2sid(ens_list_joined))
                curr_result = calculate("ENS")

                if curr_result < best_so_far:  # so dropping this star makes the result better
                    worst_ens_star = ens
                    best_so_far = curr_result
                    improved = 1

            # It was not possible to improve the optimisation by removing any star so this is optimal
            if improved == 0:
                opt_done = True
                ensemble_list_str.set(sno2sid(curr_ens_list))
                curr_result = calculate("ENS")  # this run is with the optimised ensemble
                run_status_str.set("Optimisation done: " + str(round(curr_result)))
                f_opt.close()
                finish()
                return
            else:
                remain_ens.remove(worst_ens_star)

        ensemble_list_str.set(sno2sid(ensemble_list))
        opt_done = True
        calculate("ENS")  # optimisation has probably failed - run with remaining ensemble
        run_status_str.set("Optimisation doubtful: " + sno2sid(ensemble_list))
        f_opt.close()

    else:  # this is a standard run without optimisation
        orig_ensemble = ensemble_list
        ensemble_list_str.set(sno2sid(comp_star_no))
        calculate("STD")  # run to get standard photometry
        opt_done = True
        ensemble_list_str.set(sno2sid(orig_ensemble))  # then re-run with the original ensemble
        calculate("ENS")

    finish()


################################################################################################
#
# The main body of the script which is called by the 'controller' with different ensemble lists
#
################################################################################################


def calculate(run_type):
    """The main body of the MagPy script
    run_type can be STD for standard photometry, ENS for ensemble.
    return value is the optimisation criteria for ensemble photometry, dErr/(std phot Err)^2"""

    global run_status_str, comp_star_no, comp_star_mag, comp_star_mag_str
    global in_path, in_filename_template
    global ensemble_list, ensemble_list_str, obj_id, chart, chart_str, c_name, c_name_str
    global filterband, group, group_str, notes, notes_str, obscode, obscode_str, software, software_str
    global scope_dia, scope_dia_str, exp_time, exp_time_str, ver_no, param_path, param_filename, num_stars
    global std_phot_err, mainframe, better_stars, ingress, egress, ingress_str, egress_str, date_type
    global instr_concat_lines, opt_done, time_list_std, mag_list_std, err_list_std
    global time_list_ens, mag_list_ens, err_list_ens

    run_status_str.set("Running")

    # read in the values input to the GUI

    comp_star_mag = comp_star_mag_str.get()

    ensemble_list = sid2sno(ensemble_list_str.get())
    chart = chart_str.get()
    c_name = c_name_str.get()
    group = group_str.get()
    notes = notes_str.get()
    obscode = obscode_str.get()
    software = software_str.get()
    scope_dia = scope_dia_str.get()
    exp_time = exp_time_str.get()
    ingress = ingress_str.get()
    egress = egress_str.get()

    # write the parameters for this run to the parameters file, and to an INFO file that enables
    # the run to be re-done

    # create an empty list with the required no of members so that the following assignments can be done

    f_out_lines = []
    for idx in range(23):
        f_out_lines.append("")

    f_out_lines[0] = "MagPy " + ver_no
    f_out_lines[1] = "Following data are COMP star, COMP mag, CHK star, CHK mag, path, input filename prefix, "
    f_out_lines[2] = "and stars included in the ensemble."
    f_out_lines[3] = ""
    f_out_lines[4] = comp_star_no
    f_out_lines[5] = comp_star_mag
    f_out_lines[6] = ""  # used to be chk_star_no
    f_out_lines[7] = ""  # used to be chk_star_mag
    f_out_lines[8] = in_path
    f_out_lines[9] = in_filename_template
    f_out_lines[10] = sno2sid(ensemble_list)
    f_out_lines[11] = obj_id
    f_out_lines[12] = chart
    f_out_lines[13] = c_name
    f_out_lines[14] = filterband
    f_out_lines[15] = group
    f_out_lines[16] = notes
    f_out_lines[17] = obscode
    f_out_lines[18] = software
    f_out_lines[19] = scope_dia
    f_out_lines[20] = exp_time
    f_out_lines[21] = ingress
    f_out_lines[22] = egress

    # write the INFO and parameter files

    f_out = open('INFO.txt', 'w')
    for line in f_out_lines:
        f_out.write(line + '\n')
    f_out.close()

    f_out = open(param_path + param_filename, 'w')
    for line in f_out_lines:
        f_out.write(line + '\n')
    f_out.close()

    #################################################################################################

    ###########################################################################################
    #
    # concatenate all input instr mag files into one list of lines - instr_concat_lines
    #
    ###########################################################################################

    prev_date_type = "UNDEF"  # this marks if the previous input file had HJD or JD
    instr_concat_lines = []  # will hold all lines from all input Muni files

    in_file_list = glob.glob(in_filename_template + "*.txt")
    if in_file_list == []:
        run_status_str.set("Error: No input files found.")
        return

    for f in in_file_list:
        f_in_lines = open(f).readlines()
        first_line = f_in_lines.pop(0)  # reads and removes the first line
        second_line = f_in_lines.pop(0)  # and the 2nd line, which is discarded

        # check that the first line has the expected elements

        line_elements = first_line.split()  # splits the line using any whitespace as separator
        if line_elements[-1] != "AIRMASS":
            sys.exit('Error: Input file does not have AIRMASS: ' + f)
        num_stars = (len(line_elements) - 2) / 2

        if line_elements[0] == "JDHEL":
            date_type = "HJD"
        elif line_elements[0] == "BJD":
            date_type = "BJD"
        else:
            sys.exit('Error: BJD, or JDHEL expected: ' + f)

        # checks that each input file has the same date type

        if date_type != prev_date_type and prev_date_type != "UNDEF":
            sys.exit('Error: Inconsistent date types')
        prev_date_type = date_type

        # all's well so write the concatenated data to file

        for line in f_in_lines:
            instr_concat_lines.append(line)




    #################################################################################################
    #
    # Using the concatenated list of all input lines calculate the mean reported errors of each of the
    # stars.  This is used to calculate the weighting to be given to each star in the ensemble.
    #
    #################################################################################################

    instr_mag_err_sum = []  # these lists will hold data for each measured star
    num_valid_instr_mag = []  # such that the index of the list is the star number
    avg_instr_mag_err = []

    for idx in range(int(num_stars) + 1):
        instr_mag_err_sum.append(0.0)
        num_valid_instr_mag.append(0)
        avg_instr_mag_err.append(0.0)  # initialise with all zeroes

    for line in instr_concat_lines:
        line_elements = line.split()
        for idx in range(int(num_stars)):
            curr_star_err = float(line_elements[(idx + 1) * 2])
            if curr_star_err < 9:  # only count valid observations

                instr_mag_err_sum[idx + 1] += curr_star_err
                num_valid_instr_mag[idx + 1] += 1

    for idx in range(int(num_stars)):
        avg_instr_mag_err[idx + 1] = instr_mag_err_sum[idx + 1] / num_valid_instr_mag[idx + 1]


    ##############################################################################################
    #
    # calculate the weighted average mag of the ensemble stars for each timestep.
    # Then subtract that from the instr mag of each star.
    # Write results to reduced_mag_lines
    #
    ##############################################################################################

    # get the list of stars in the ensemble and calculate the divisor for the weighted avg mag

    ensemble_list_list = ensemble_list.split(',')  # the list of ensemble stars is comma delimited
    num_ensemble_stars = len(ensemble_list_list)

    for idx in range(num_ensemble_stars):
        ensemble_list_list[idx] = int(ensemble_list_list[idx])  # convert from string to integer

    if int(comp_star_no) in ensemble_list_list:
        aavso_report_invalid = True  # AAVSO ensemble format does not allow COMP to be one of the ensemble
    else:
        aavso_report_invalid = False

    wt_avg_mean_mag_divisor = 0.0

    for ens in ensemble_list_list:
        wt_avg_mean_mag_divisor += 1.0 / avg_instr_mag_err[ens]


    # for each line in instr_concat_lines calculate the weighted average of the ensemble
    # then calculate the mag of each star with respect to this weighted average and add to reduced_mag_lines

    valid_ens_obs = 0  # counter for the number of valid ensemble observations
    valid_comp_obs = 0  # counter for no of lines with valid COMP mag
    reduced_comp_mag_sum = 0  # used to calculate the average reduced COMP mag, to calc the needed correction

    reduced_mag_lines = []

    for line in instr_concat_lines:
        valid_line = 1
        line_elements = line.split()
        time = line_elements[0]
        airmass = float(line_elements[-1])
        scint_err = Scint_Err(airmass, float(exp_time))

        sum_instr_mag_over_error = 0.0
        sum_recip_err_squared = 0.0

        for ens in ensemble_list_list:  # for each star in the ensemble
            mag = float(line_elements[ens * 2 - 1])
            err = math.sqrt(float(line_elements[ens * 2]) ** 2.0 + scint_err ** 2.0)  # introduce scintillation error
            if mag > 50:
                valid_line = 0  # any bad ensemble star renders the whole timestep invalid
                break
            else:
                sum_instr_mag_over_error += mag / avg_instr_mag_err[ens]
                sum_recip_err_squared += 1.0 / (err ** 2.0)

        if valid_line == 1:
            ensemble_wt_avg_mag = sum_instr_mag_over_error / wt_avg_mean_mag_divisor
            ensemble_err = math.sqrt(1 / sum_recip_err_squared)

            out_line = time  # starts each line that will be held in reduced_mag_lines
            for star in range(1, int(num_stars) + 1):
                mag = float(line_elements[star * 2 - 1])
                err = float(line_elements[star * 2])
                out_line += " " + str(mag - ensemble_wt_avg_mag)
                out_line += " " + str(math.sqrt(err ** 2 + ensemble_err ** 2))

                if star == int(comp_star_no) and err < 9:  # count only valid COMP obs
                    reduced_comp_mag_sum += (mag - ensemble_wt_avg_mag)
                    valid_comp_obs += 1
            out_line += " " + str(airmass)
            reduced_mag_lines.append(out_line)
            valid_ens_obs += 1

    reduced_comp_avg_mag = reduced_comp_mag_sum / valid_comp_obs


    ##########################################################################################################
    #
    # This section shifts each reduced mag to the standard scale
    # with respect to the average of the reduced COMP mags, ie after shifting, average COMP mag equals
    # catalog COMP mag.
    # This section also does the following;
    # 1.  writes either STD_PHOT_STAR_x.txt or ENSEMBLE_STAR_x.txt files to each results subdirectory
    # 2.  writes the AAVSO_obj_id.txt file
    # 3.  calculates the mean and std dev of each star and records them in the STD_PHOT or ENSEMBLE_STATS.txt file.
    #
    #########################################################################################################

    mag_corr = float(comp_star_mag) - reduced_comp_avg_mag

    if run_type == "STD":  # sets different filenames depending on the run type
        ftype = "STD_PHOT"
        std_phot_err = []  # will hold the std photometry error for each star
    elif run_type == "ENS":
        ftype = "ENSEMBLE"
        ens_phot_err = []  # will hold the ensemble photometry error for each star
    else:
        run_status_str.set("Unexpected Run Type")
        return

    # open the AAVSO output file and initialise - only if this is a results run, ie opt_done is True
    if opt_done:
        if replace_hyphen.get():
            aavso_obj_id = re.sub('[-.]', ' ', obj_id)  # replace any dash in obj name with space
        else:
            aavso_obj_id = obj_id

        if not aavso_report_invalid:
            AAVSO_filename = "AAVSO_" + obj_id + ".txt"
        else:
            AAVSO_filename = "INVALID_AAVSO_" + obj_id + ".txt"

        aavso_file = open(AAVSO_filename, 'w')
        aavso_file.write("#TYPE=EXTENDED\n")
        aavso_file.write("#OBSCODE=" + obscode + "\n")
        aavso_file.write("#SOFTWARE=" + software + "\n")
        aavso_file.write("#DELIM=,\n")
        aavso_file.write("#DATE=" + date_type + "\n")
        aavso_file.write("#OBSTYPE=CCD\n")
        aavso_file.write("#NAME,DATE,MAG,MERR,FILT,TRANS,MTYPE,CNAME,CMAG,KNAME,KMAG,AMASS,GROUP,CHART,NOTES\n")

    # make lists to hold time, mag, err data for each star, and for std and ens runs

    if run_type == 'STD':
        time_list_std = [[]]  # this will be a list of lists. This is the first, empty, list
        mag_list_std = [[]]
        err_list_std = [[]]
        for star in range(1, int(num_stars) + 1):
            time_list_std.append([])
            mag_list_std.append([])
            err_list_std.append([])  # the index will be star no. Index 0 will always be an empty list

    if run_type == 'ENS':
        time_list_ens = [[]]
        mag_list_ens = [[]]
        err_list_ens = [[]]
        for star in range(1, int(num_stars) + 1):
            time_list_ens.append([])
            mag_list_ens.append([])
            err_list_ens.append([])

    for line in reduced_mag_lines:
        star1_invalid = 0  # reset this flag which marks if the Star 1 measure on this line is invalid
        line_elements = line.split()
        time = line_elements[0]
        airmass = line_elements[-1]

        for star in range(1, int(num_stars) + 1):
            mag = float(line_elements[star * 2 - 1]) + mag_corr
            err = float(line_elements[star * 2])

            if err < 9.0:  # Muniwin uses 9.999 to denote invalid measure
                if star == 1:
                    obj_mag = round(mag, 4)
                    obj_err = round(err, 4)  # capture this data for the AAVSO file
                else:
                    if star == int(comp_star_no):
                        c_mag = round(mag, 4)  # this is AAVSO KMAG

                if run_type == 'STD':
                    time_list_std[star].append(time)  # string
                    mag_list_std[star].append(mag)  # float
                    err_list_std[star].append(err)  # float
                elif run_type == 'ENS':
                    time_list_ens[star].append(time)  # string
                    mag_list_ens[star].append(mag)  # float
                    err_list_ens[star].append(err)  # float
            else:
                if star == 1:
                    star1_invalid = 1  # marks not to write this line to the AAVSO file
                if star == int(comp_star_no):
                    star1_invalid = 1  # marks not to write this line to the AAVSO file

        if star1_invalid != 1 and opt_done:
            out_line1 = aavso_obj_id + "," + time + "," + str(obj_mag) + "," + str(obj_err) + "," + filterband
            out_line2 = out_line1 + ",NO,STD,ENSEMBLE,NA," + c_name + "," + str(c_mag) + "," + airmass
            out_line3 = out_line2 + "," + group + "," + chart + "," + notes + "\n"
            aavso_file.write(out_line3)

    if opt_done:
        aavso_file.close()

    #####################################################################################################
    # this section calculates the mean ensemble mag of each star, and the std dev

    opt_criteria = 0.0  # this will become the return value in a optimisation run

    add_line1 = ""
    add_line2 = ""

    if run_type == "ENS":  # info to add to the ENSEMBLE_STATS.txt file
        add_line1 = "  For STD, then ENS measures."
        add_line2 = "\t\tMean\tSD\tNo. Obs"

        # the optimisation criteria will only use stars that have better than average scatters
        # this section makes a list of these stars.
        # The COMP star will be assigned the average scatter - this is pretty arbitrary

        sum_scatter = 0.0
        better_stars = []
        bad_stars = 0  # count no of stars with invalid no of observations
        for star in range(2, int(num_stars) + 1):  # exclude Star 1 from the count
            if star != int(comp_star_no):  # also exclude COMP, which has scatter = 0
                if len(mag_list_std[star]) <= 1: # ie a bad star
                    bad_stars += 1
                else:
                    sum_scatter += np.nanstd(mag_list_std[star], ddof=1)

        avg_scatter = sum_scatter / (num_stars - 2 - bad_stars)

        for star in range(2, int(num_stars) + 1):  # exclude Star 1 from the count
            if np.nanstd(mag_list_std[star], ddof=1) <= avg_scatter:
                better_stars.append(star)

    ens_stats_file = open(ftype + "_STATS.txt", 'w')
    ens_stats_file.write("MagPy " + ver_no + "\n")
    ens_stats_file.write("Following data are Star label, Star id, Mean mag, Std Dev of the magnitudes, "
                         "and number of obs per star." + add_line1 + "\n")
    ens_stats_file.write("\n")
    ens_stats_file.write("COMP is star id " + str(star_ids[2 - 1]) + ", mag " + comp_star_mag + "\n")
    ens_stats_file.write("Ensemble stars: " + sno2sid(ensemble_list) + "\n")
    if detrend.get():
        ens_stats_file.write("Detrended ensemble stars used.\n")

    ens_stats_file.write("\n")
    ens_stats_file.write("Label\tID\tMean\tSD\tNo. Obs" + add_line2 + "\n")
    ens_stats_file.write("\n")
    ens_stats_file.write("\n")

    for star in range(1, int(num_stars) + 1):
        add_ens_data = ""
        av_mag_std = np.nanmean(mag_list_std[star])
        sd_std = np.nanstd(mag_list_std[star], ddof=1)
        num_pt_std = len(mag_list_std[star])

        sid = star_ids[star - 1]

        if run_type == "ENS":
            av_mag_ens = np.nanmean(mag_list_ens[star])
            sd_ens = np.nanstd(mag_list_ens[star], ddof=1)
            num_pt_ens = len(mag_list_ens[star])

            add_ens_data = "\t\t" + str(round(av_mag_ens,4)) + "\t" + str(round(sd_ens,4)) \
                           + "\t" + str(num_pt_ens)

        if run_type == "STD":
            std_phot_err.append(sd_std)
        elif run_type == "ENS":
            ens_phot_err.append(sd_ens)  # record the errors for use in optimisation
            if star == int(comp_star_no):
                std_phot_err[star - 1] = avg_scatter / 2  # arbitrary, otherwise this would be zero
            if star in better_stars:
                opt_criteria += (ens_phot_err[star - 1] - std_phot_err[star - 1]) / (std_phot_err[star - 1] ** 2.0)

        ens_stats_file.write(star_labels[star - 1] + "\t" + str(sid) + "\t" + str(round(av_mag_std, 4)) +
                             "\t" + str(round(sd_std, 4)) + "\t" + str(num_pt_std) + add_ens_data + "\n")
    ens_stats_file.close()

    run_status_str.set("Opt Criteria: " + str(round(opt_criteria)))
    mainframe.update_idletasks()  # forces update of the status display
    return opt_criteria

    ###################################################################################################
    #
    # End calculate
    #
    ###################################################################################################


########################################################################################
#
#  Start of script proper
#
########################################################################################


# read config file

homedir = os.path.expanduser('~')       # find the home directory

config = open(os.path.join(homedir, 'bin', 'pest_pipeline_config.txt'), 'r').readlines()
config = [x.strip() for x in config]  # remove leading and trailing whitespaces

for i in range(len(config)):
    if config[i] != '':
        if config[i][0] == '#':
            config[i] = ''      # set all comment lines to blank

param_path = get_conf_var('magpy_param_path', config, 'NA')
observatory = get_conf_var('observatory', config, 'NA')
telescope = get_conf_var('telescope', config, 'NA')
camera = get_conf_var('camera', config, 'NA')
tess_user = get_conf_var('tess_user', config, 'NA')

scale = float(get_conf_var('scale', config, 'NA'))        # arcsec per pixel
aavso_code = get_conf_var('aavso_obs_code', config, 'NA')
signature = get_conf_var('signature', config, 'NA')

param_filename = "param_MagPy.txt"
in_filename_template = 'Muni'  # there must exist Muni files. If the 'use detrend' is ticked will check later
# for the det-Muni files

##########################################################################################################
# Read the select_labels.txt file so as to use star-ids in the data products
# If it's not there look for it in other observations of the same target
##########################################################################################################
if not os.path.exists('select_labels.txt'):
    labels_file_list = glob.glob(os.path.join(os.pardir,'*','select_labels.txt'))
    if len(labels_file_list) > 0:
        latest_file = max(labels_file_list, key=os.path.getctime)
        print('Found existing select_labels.txt file and copying that over: ' + latest_file)
        copyfile(latest_file, 'select_labels.txt')

if os.path.exists('select_labels.txt'):
    # load the file into a dataframe, skipping first 2 rows, using any whitespace as separator
    # the engine='python' is needed because the c engine cannot parse the regex "\s+"

    label_frame = pd.read_csv('select_labels.txt', sep="\s+", skiprows=2, header=None, engine='python')

    label_frame.columns = ['label', 'id']
    label_frame.set_index('label', inplace=True)

    star_ids = label_frame['id'].values  # array with all star ids
    star_labels = label_frame.index.values
    var_id = label_frame.loc['var']['id']  # star id of VAR

else:
    sys.exit('ERROR:  select_labels.txt, produced by photcorr, not found.\n')

#########################################################################################
# Find latest INFO.txt file found under the same target directory
# and give user the choice of whether to overwrite the MagPy params file

info_file_list = glob.glob(os.path.join(os.pardir,'*','INFO.txt'))
if len(info_file_list) > 0:
    latest_file = max(info_file_list, key=os.path.getctime)
    print('Found existing INFO.txt file: ' + latest_file)
    in_lines = open(latest_file).readlines()
    info_obj_id = in_lines[11].strip()

    #if not os.path.exists('INFO.txt'):
    #    copyfile(latest_file, os.path.join(in_path, 'INFO.txt'))

    if os.path.exists(os.path.join(param_path, 'param_MagPy.txt')):  # find out what target is in the MagPy file
        in_lines = open(os.path.join(param_path, 'param_MagPy.txt')).readlines()
        file_obj_id = in_lines[11].strip()
        if file_obj_id == info_obj_id:
            print('Target in existing INFO.txt is consistent with MagPy parameters file: ' + info_obj_id)
        else:
            print('Current MagPy parameters file has a different target: ' + file_obj_id)
            overwrite_param = input('Overwrite MagPy parameters? (Y/y for Yes, Enter or anything else for No) ').strip()
            if overwrite_param == 'Y' or overwrite_param == 'y':
                copyfile(latest_file, os.path.join(param_path, 'param_MagPy.txt'))


#############################################################################################
#
# Declare variables and starting values, Read input parameters from the MAGPY_PARAM.txt file
#
#############################################################################################


run_status = "Waiting for run..."
block_size = '270'  # default 270s

if os.path.exists(param_path + param_filename):
    f_in_lines = open(param_path + param_filename).readlines()  # read all lines in this file into a list
    comp_star_no = '2'                      # a legacy line left in so as not to disrupt the params file
    comp_star_mag = f_in_lines[5].rstrip()
    chk_star_no = ''          # a legacy line left in so as not to disrupt the params file
    chk_star_mag = ''          # a legacy line left in so as not to disrupt the params file
    #in_path = f_in_lines[8].rstrip()               # a legacy line left in so as not to disrupt the params file
    #in_filename_template = 'Muni'                   # legacy item
    ensemble_list = sid2sno(f_in_lines[10].rstrip())
    obj_id = f_in_lines[11].rstrip()
    chart = f_in_lines[12].rstrip()
    c_name = f_in_lines[13].rstrip()
    #filterband = f_in_lines[14].rstrip()
    group = f_in_lines[15].rstrip()
    notes = f_in_lines[16].rstrip()
    obscode = f_in_lines[17].rstrip()
    software = f_in_lines[18].rstrip()
    scope_dia = f_in_lines[19].rstrip()
    exp_time = f_in_lines[20].rstrip()

    if len(f_in_lines) > 21:                    # to maintain compatibility with old versions of the param file
        ingress = f_in_lines[21].rstrip()
        egress = f_in_lines[22].rstrip()
    else:
        ingress = '0'
        egress = '0'

else:  # on an initial run use these defaults
    comp_star_no = '2'
    comp_star_mag = '11'
    chk_star_no = ''
    chk_star_mag = ''
    #in_path = ""                            # a legacy line, the working directory will be used
    #in_filename_template = "Muni"
    ensemble_list = " "
    obj_id = "NA"
    chart = "NA"
    c_name = "NA"
    #filterband = "Rc"
    group = "1"
    notes = "NA"
    obscode = aavso_code
    software = "C-Munipack V2.1.25"
    scope_dia = '305'
    exp_time = '120'
    ingress = '0'
    egress = '0'

# if PESTphot was run before MagPy, there will be a file with info on COMP mag, ingress/egress, and max_dmag
mt_populated = False  # marks that the info below has been read from file
if os.path.exists('PESTphot_inputs.txt'):
    mt_lines = open('PESTphot_inputs.txt').readlines()
    comp_star_mag = mt_lines[0].strip()
    ingress = mt_lines[1].strip()
    egress = mt_lines[2].strip()
    mt_populated = True


##################################################################################################
#
# Sets the in_path to the current directory, and obtains obj_id and obs_date from the path
#
##################################################################################################

in_path = os.getcwd()

obj_id = os.path.basename(os.path.dirname(in_path))
obs_date = os.path.basename(in_path)

obs_date_strip = re.sub('[-.]', '', obs_date)  # strips out any dash or dot in the directory name


#########################################################################################
# make a subdirectory for working files as well as individual star data files

magpy_dir = os.path.relpath('magpy_files', start=os.curdir)

if not os.path.exists(magpy_dir):
    os.mkdir(magpy_dir)


###########################################################################################################
# check if this will be a multi input file run
# find the filter band

multi = False

in_file_list = glob.glob(in_filename_template + '*.txt')    # find input files
if in_file_list == []:
    sys.exit("Error: No file/s named '{:s}*.txt' found".format(in_filename_template))

if len(in_file_list) > 1:
    multi = True
    block_size = '0'  # for legibility

in_file = open(in_file_list[0])     # read just the first file found
for line in in_file:
    if re.search(r"Aperture:\s+([0-9]+),\s+Filter:\s+([\w]+),", line):        # this line has the aperture no
        sub_line = re.search(r"Aperture:\s+([0-9]+),\s+Filter:\s+([\w]+),", line)
        filterband = sub_line.group(2)

in_file.close()


##########################################################################################################
# find the photometry aperture radius and exposure time

if not os.path.exists('PESTphot_info.txt'):
    pp_file_list = glob.glob(os.path.join(os.pardir, '*', 'PESTphot_info.txt'))
    if len(pp_file_list) > 0:
        latest_file = max(pp_file_list, key=os.path.getctime)
        print('Found existing PESTphot_info.txt file: copying this: ' + latest_file)
        copyfile(latest_file, 'PESTphot_info.txt')

    pp_file_list = glob.glob(os.path.join(os.pardir, '*', 'photcorr_info.txt'))
    if len(pp_file_list) > 0:
        latest_file = max(pp_file_list, key=os.path.getctime)
        print('Found existing photcorr_info.txt file: copying and renaming this: ' + latest_file)
        copyfile(latest_file, 'PESTphot_info.txt')


if os.path.exists('PESTphot_info.txt'):
    pcinfo = open('PESTphot_info.txt').readlines()
    phot_rad = float(pcinfo[2].rsplit(None,1)[-1])  # the end of the 3rd line is the rad in pixels
    obs_dur = pcinfo[3].rsplit(None,1)[-1]  # the end of the 4th line is the observation duration in mins
    #obs_frames = pcinfo[4].rsplit(None,1)[-1]  # the end of the 5th line is the no of frames

    fwhm = pcinfo[6].rsplit(None,2)[-2]  # the 2nd last item of the 7th line is the FWHM in arcsec

    #filter = pcinfo[5].rsplit(None,1)[-1]  # the end of the 6th line is the filter
    exp_time = pcinfo[8].rsplit(None,1)[-1]  # the end of the 9th line is the exp time in s

    obs_dur = str(round(float(obs_dur)))  # round to zero decimal place
    fwhm = str(round(float(fwhm),1))  # round to one decimal place


else:
    print('PESTphot_info.txt not present. Requesting input...')
    phot_rad = float(input('Enter the photometry aperture radius in pix: '))
    obs_dur = '0'  # these are used only in reporting, so just set to 0 if PESTphot_info.txt does not exist
    #obs_frames = '0'
    fwhm = '0'
    #filter = input('Enter the filter: ')
    exp_time = input('Exposure time in s: ')

phot_rad_arcsec = phot_rad*scale

# the standard part of key output filenames
std_name = obj_id + '_' + obs_date_strip + '_' + observatory + '_' + filterband


##########################################################################################
#
# This sets up the GUI
#
##########################################################################################

root = Tk()
root.style = ttk.Style()
#('clam', 'alt', 'default', 'classic')
root.style.theme_use("clam")
#root.style.configure('TLabel', font='arial 10 bold')
#root.style.configure('TEntry', font='verdana 10 bold')
#root.style.configure('TCheckbutton', font='verdana 10 bold')



root.title(script_name + ver_no)

mainframe = tkinter.ttk.Frame(root, padding="10 20 10 20")
mainframe.grid(column=0, row=0, sticky=(N, W, E, S))

for num in range(30):
    mainframe.rowconfigure(num, pad=6, weight=1)
for num in range(5):
    mainframe.columnconfigure(num, pad=10, weight=1)


#comp_star_str = StringVar()
#comp_star_str.set(comp_star_no)
#tkinter.ttk.Label(mainframe, text="COMP (1 to n): ").grid(column=1, row=1, sticky=W, columnspan=2)
#comp_star_entry = tkinter.ttk.Entry(mainframe, width=7, textvariable=comp_star_str)
#comp_star_entry.grid(column=3, row=1, sticky=W)

comp_star_mag_str = StringVar()
comp_star_mag_str.set(comp_star_mag)
tkinter.ttk.Label(mainframe, text="Mag of COMP: ").grid(column=1, row=1, sticky=W, columnspan=2)
comp_star_mag_entry = tkinter.ttk.Entry(mainframe, width=7, textvariable=comp_star_mag_str)
comp_star_mag_entry.grid(column=3, row=1, sticky=W)

#chk_star_str = StringVar()
#chk_star_str.set(chk_star_no)
#tkinter.ttk.Label(mainframe, text="CHK (1 to n): ").grid(column=1, row=3, sticky=W, columnspan=2)
#chk_star_entry = tkinter.ttk.Entry(mainframe, width=7, textvariable=chk_star_str)
#chk_star_entry.grid(column=3, row=3, sticky=W)

#chk_star_mag_str = StringVar()
#chk_star_mag_str.set(chk_star_mag)
#tkinter.ttk.Label(mainframe, text="Mag of CHK: ").grid(column=1, row=4, sticky=W, columnspan=2)
#chk_star_mag_entry = tkinter.ttk.Entry(mainframe, width=7, textvariable=chk_star_mag_str)
#chk_star_mag_entry.grid(column=3, row=4, sticky=W)

c_name_str = StringVar()
c_name_str.set(c_name)
tkinter.ttk.Label(mainframe, text="COMP Label: ").grid(column=4, row=1, sticky=W)
c_name_entry = tkinter.ttk.Entry(mainframe, width=7, textvariable=c_name_str)
c_name_entry.grid(column=5, row=1, sticky=W)

tkinter.ttk.Label(mainframe, text="COMP places mags on std scale,").grid(column=4, row=2, sticky=W, columnspan=2)
tkinter.ttk.Label(mainframe, text="in AAVSO Ensemble = CHK.").grid(column=4, row=3, sticky=W, columnspan=2)

spacer1 = tkinter.ttk.Frame(mainframe, height=10).grid(column=1, row= 5, columnspan=5)

tkinter.ttk.Label(mainframe, text="Input below are BJD/HJD - 2450000. Enter 0 if no line needed.")\
    .grid(column=1, row=6, sticky=W, columnspan=5)

ingress_str = StringVar()
ingress_str.set(ingress)
tkinter.ttk.Label(mainframe, text="Ingress BJD/HJD: ").grid(column=1, row=7, sticky=W, columnspan=2)
ingress_entry = tkinter.ttk.Entry(mainframe, width=8, textvariable=ingress_str)
ingress_entry.grid(column=3, row=7, sticky=W)

egress_str = StringVar()
egress_str.set(egress)
tkinter.ttk.Label(mainframe, text="Egress: ").grid(column=4, row=7, sticky=W, columnspan=2)
egress_entry = tkinter.ttk.Entry(mainframe, width=8, textvariable=egress_str)
egress_entry.grid(column=5, row=7, sticky=W)

spacer11 = tkinter.ttk.Frame(mainframe, height=10).grid(column=1, row=8, columnspan=5)

#in_path_str = StringVar()
#in_path_str.set(in_path)
#in_path_entry = tkinter.ttk.Entry(mainframe, textvariable=in_path_str).grid(column=1, row=9, columnspan=5, sticky=(E, W))

tkinter.ttk.Label(mainframe, text='Working dir: .../'+obj_id+'/'+obs_date).grid(column=1, row=9, sticky=W, columnspan=5)

block_str = StringVar()
block_str.set(block_size)
tkinter.ttk.Label(mainframe, text="Block size (s): ").grid(column=1, row=11, sticky=W, columnspan=2)
block_entry = tkinter.ttk.Entry(mainframe, width=8, textvariable=block_str)
block_entry.grid(column=3, row=11, sticky=W)
tkinter.ttk.Label(mainframe, text="enter 0 if block plotting not needed.").grid(column=4, row=11, sticky=W, columnspan=2)

detrend = IntVar()
tkinter.ttk.Checkbutton(mainframe, text="  Use detrended ens? ", variable=detrend).grid(column=5, row=16,
                                                                               columnspan=2, sticky=W)

#in_filename_template_str = StringVar()
#in_filename_template_str.set(in_filename_template)
#in_filename_template_entry = tkinter.ttk.Entry(mainframe, width=15, textvariable=in_filename_template_str).grid(column=3,row=8, columnspan=5, sticky=W)

spacer2 = tkinter.ttk.Frame(mainframe, height=10).grid(column=1, row=11, columnspan=5)

spacer3 = tkinter.ttk.Frame(mainframe, height=10).grid(column=1, row=12, columnspan=5)

tkinter.ttk.Label(mainframe, text="List of stars in ensemble, comma separated:").grid(column=1, row=13, sticky=W,columnspan=5)
ensemble_list_str = StringVar()
ensemble_list_str.set(sno2sid(ensemble_list))
ensemble_list_entry = tkinter.ttk.Entry(mainframe, textvariable=ensemble_list_str).grid(column=1, row=14, columnspan=5, \
                                                                                sticky=(E, W))
tkinter.ttk.Label(mainframe, text="AAVSO Report will be marked INVALID if ensemble uses COMP.").grid(column=1, row=15, sticky=W,\
                                                                                         columnspan=5)

optimise = IntVar()
tkinter.ttk.Checkbutton(mainframe, text="  Optimise ensemble? ", variable=optimise).grid(column=2, row=16,
                                                                                 columnspan=3, sticky=W)

#spacer4 = tkinter.ttk.Frame(mainframe, height=10).grid(column=1, row=17, columnspan=5)
tkinter.ttk.Separator(mainframe, orient=HORIZONTAL).grid(column=1, row=17, sticky=(E,W), columnspan=5)
spacer5 = tkinter.ttk.Frame(mainframe, height=10).grid(column=1, row=18, columnspan=5)

replace_hyphen = IntVar()
tkinter.ttk.Checkbutton(mainframe, text="  In AAVSO report, replace hyphen in Obj ID with space? ",
                        variable=replace_hyphen).grid(column=1, row=19, columnspan=5, sticky=W)

#obj_id_str = StringVar()
#obj_id_str.set(obj_id)
tkinter.ttk.Label(mainframe, text="Obj ID: ").grid(column=1, row=20, sticky=W)
#obj_id_entry = tkinter.ttk.Entry(mainframe, width=15, textvariable=obj_id_str)
#obj_id_entry.grid(column=2, row=20, sticky=W, columnspan=2)
tkinter.ttk.Label(mainframe, text=obj_id).grid(column=2, row=20, sticky=W, columnspan=3)


chart_str = StringVar()
chart_str.set(chart)
tkinter.ttk.Label(mainframe, text="Chart ID: ").grid(column=4, row=20, sticky=W)
chart_entry = tkinter.ttk.Entry(mainframe, width=18, textvariable=chart_str)
chart_entry.grid(column=5, row=20, sticky=W)

#filterband_str = StringVar()
#filterband_str.set(filterband)
tkinter.ttk.Label(mainframe, text="Filter: ").grid(column=1, row=21, sticky=W)
#filterband_entry = tkinter.ttk.Entry(mainframe, width=10, textvariable=filterband_str)
#filterband_entry.grid(column=2, row=21, sticky=W, columnspan=2)
tkinter.ttk.Label(mainframe, text=filterband).grid(column=2, row=21, sticky=W, columnspan=3)


group_str = StringVar()
group_str.set(group)
tkinter.ttk.Label(mainframe, text="Group: ").grid(column=4, row=21, sticky=W)
group_entry = tkinter.ttk.Entry(mainframe, width=18, textvariable=group_str)
group_entry.grid(column=5, row=21, sticky=W)

notes_str = StringVar()
notes_str.set(notes)
tkinter.ttk.Label(mainframe, text="Notes: ").grid(column=1, row=22, sticky=W)
notes_entry = tkinter.ttk.Entry(mainframe, width=30, textvariable=notes_str)
notes_entry.grid(column=2, row=22, sticky=(E,W), columnspan=5)

obscode_str = StringVar()
obscode_str.set(obscode)
tkinter.ttk.Label(mainframe, text="Obs Code: ").grid(column=1, row=23, sticky=W)
obscode_entry = tkinter.ttk.Entry(mainframe, width=10, textvariable=obscode_str)
obscode_entry.grid(column=2, row=23, sticky=W, columnspan=2)

software_str = StringVar()
software_str.set(software)
tkinter.ttk.Label(mainframe, text="Software: ").grid(column=4, row=23, sticky=W)
software_entry = tkinter.ttk.Entry(mainframe, width=18, textvariable=software_str)
software_entry.grid(column=5, row=23, sticky=W)

spacer6 = tkinter.ttk.Frame(mainframe, height=10).grid(column=1, row=24, columnspan=5)
tkinter.ttk.Separator(mainframe, orient=HORIZONTAL).grid(column=1, row=25, sticky=(E,W), columnspan=5)
spacer7 = tkinter.ttk.Frame(mainframe, height=10).grid(column=1, row=26, columnspan=5)

scope_dia_str = StringVar()
scope_dia_str.set(scope_dia)
tkinter.ttk.Label(mainframe, text="Scope Dia/mm: ").grid(column=1, row=27, sticky=W, columnspan=2)
scope_dia_entry = tkinter.ttk.Entry(mainframe, width=7, textvariable=scope_dia_str)
scope_dia_entry.grid(column=3, row=27, sticky=W)

exp_time_str = StringVar()
exp_time_str.set(exp_time)
tkinter.ttk.Label(mainframe, text="Exp Time/s: ").grid(column=4, row=27, sticky=W)
exp_time_entry = tkinter.ttk.Entry(mainframe, width=7, textvariable=exp_time_str)
exp_time_entry.grid(column=5, row=27, sticky=W)

spacer8 = tkinter.ttk.Frame(mainframe, height=20).grid(column=1, row=28, columnspan=5)
spacer9 = tkinter.ttk.Frame(mainframe, height=50).grid(column=1, row=31, columnspan=5)

#tkinter.ttk.Button(mainframe, text="Browse", width=12, command=in_dirname).grid(column=3, row=6, sticky=W)
#tkinter.ttk.Button(mainframe, text="Cur Dir", width=12, command=cur_dirname).grid(column=4, row=6, sticky=W, columnspan=2)
tkinter.ttk.Button(mainframe, text="RUN", width=10, command=controller).grid(column=3, row=31, sticky=(N,S,W))
tkinter.ttk.Button(mainframe, text="CLOSE", width=10, command=quit).grid(column=4, row=31, sticky=(N,S,W))

spacer10 = tkinter.ttk.Frame(mainframe, height=20).grid(column=1, row=32, columnspan=5)

run_status_str = StringVar()
run_status_str.set(run_status)
tkinter.ttk.Label(mainframe, text="Status: ").grid(column=1, row=33, sticky=W, columnspan=2)
status_display = tkinter.ttk.Label(mainframe, foreground='red',
                           textvariable=run_status_str).grid(column=3, row=33, sticky=W, columnspan=4)

comp_star_mag_entry.focus()
root.bind('<Return>', controller)

root.mainloop()


######################################################################################################
#
# End of the GUI section
#
######################################################################################################

