#! /usr/bin/env python3

# PESTphot
# Author:         TG Tan   www.pestobservatory.com
#
# V0.0      29/4/19     Draft version
# V1.0      9/5/19      First working version
# V1.1      10/5/19     Write out max_dmag for magpy to pick up
# V1.2      22/5/19     if INFO.txt file is present, use that for the COMP info
# V1.3      23/5/19     remove unnecessary plate-solve files, bug fix on ingress/egress values
# V1.4      23/5/19     add config option to turn off detrending and associated plotting,
#                       limit outlier-star plots to proximate ones
# V1.5      24/5/19     change filename of photcorr_info.txt to PESTphot_info.txt.
#                       Write data to facilitate adding TESS timeseries (moved from MagPy)
#                       Read date_type from config file, rather than as a command line switch
#                       The aperture number to be used is optional on command line - input will be requested if missing
#                       Write detrend parameters used to to PESTphot_info.txt.
# V1.6      26/5/19     Look for and copy over INFO.txt file from previous obs, copy into param_MagPy.txt as well
# V1.7      28/5/19     Check for consistency of apertures where previous obs of the same target have been made
#                       At input request, just press Enter to use previous aperture
# V1.8      3/6/19      Print current input values on same line as input prompts. Exit if Gaia cannot match VAR
#                       Change FWHM plot and mean to be of COMP rather than VAR - VAR is sometimes saturated
# V1.9      20/6/19     Minor: print out target name at input, bug fix - limit prox stars to those with >1 point
# V2.0      22/6/19     Remove unnecessary reading of tags, trap errors from no position data, in the XML file
# V2.1      4/7/19      Add option to skip plate-solve
# V2.2      13/7/19     Add early warning and exit if there is no catalog file
# V2.3      17/7/19     Add 'simple' mode that does not plot proximate lightcurves, detrend, or do plate-solve
# V2.4      7/9/19      Bug fix for plotting field charts when some pixels are -ve
#
# todo - box transit model
#
# Script Function:
# Given a set of photometry files (.pht) produced by C-Munipack, produce;
# 1. Un detrended and detrended instrumental mags for further ensemble processing by MagPy
# 2. Informational and diagnostic charts from the observation
# 3. Provide diagnostics for EB checks, plate solving and plotting of finder charts
#
# Syntax:
#  python3 PESTphot.py <aperture>
#
#
#
# Files produced:
# 1. Muni...txt: file for input to MagPy.  This is always un-detrended data.
# 2. magdev.txt: data corresponding to the magdev chart
# 3. ...airmass-fwhm.png: Plot, against time
# 4. ...xy-shift.png: Position of the target star on the CCD with time
# 5. ...raw-instr-mag.png: Plot for VAR and COMP
# 7. PESTphot_info.txt with info on ver, aperture, detrend params,
# 9. Light curves of outlier stars within 2.5'
# 11. select_labels.txt with selection star labels and star ids

# If plate solving is on;
# 12. proximate_star_n.png light curves for stars within 2.5'

#
# If detrending is switched on;
# 14. det_Muni...txt: Muni file as above, but for detrended data
# 15. ...detrend-data.txt: containing the data used for detrending
# 16. regression_coeffs.txt:  Regression coefficients calculated
# 17. detrend_stats.txt: Std dev of selected stars before and after detrend
# 18. Light curves of selected stars pre and post detrend (if detrend plots is on)
#
# Assumptions:
# 1. There is a <target_name>.xml file in the Catalog directory
# 2. Both target coordinates and observatory long and lat are correctly recorded in the xml
#       so that BJD times are calculated correctly.
# 3. Astropy is at least ver 2
# 4. <obs_date> is in the form YYYY-MM-DD, or YYYYMMDD
# 5. Only one VAR and one COMP star are selected.
# 6. The ~/bin/pest_pipeline_config.txt file exists
#
#


##################################################################################################
# The Gaia check sections are based on a script by John Kielkopf, University of Louisville
# The code is open source and the MIT License applies
##################################################################################################



from tkinter import *
import math, subprocess, re, os, glob
from astropy.io import fits
from astropy import wcs
from astropy import units as u
from astropy.coordinates import Angle
import pandas as pd
import numpy as np
import aplpy
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import zipfile
from shutil import copyfile

from astropy import coordinates
from astroquery.gaia import Gaia

from astropy.visualization import astropy_mpl_style
plt.style.use(astropy_mpl_style)                # use this improved style



import struct, sys
from xml.dom import minidom
from astropy import time, coordinates as coord
from astropy.stats import median_absolute_deviation
from sklearn import linear_model
from shutil import copyfile

global target_coord, obs_location, long_obj, lat_obj, init_time, tot_frames, frm

script_name = 'PESTphot by TG Tan: '
ver_no = 'Ver 2.4'
print('\n')
print(script_name, ver_no)



###########################################################################
# Function to return the Julian date for a given year, month, day, and UTC
###########################################################################

def date2jd(y0,m0,d0,u0):
    k = int(y0)
    m = int(m0)
    i = int(d0)
    utc = u0
    j1 = float(367*k)
    j2 = -1.*float( int( ( 7*(k + int((m+9)/12) ) )/4 ) )
    j3 = float(i + int(275*m/9))
    j4 = 1721013.5 + (utc/24.)
    julian = j1 + j2 + j3 +j4
    return(julian)


############################################################################
# Function to return the Julian date for a given floating point epoch
############################################################################

def jd_for_epoch(epoch):
    int_year = int(epoch)
    month  = epoch - float(int_year)
    int_month = int(month)
    day = int_month - float(int_month)
    int_day = int(day)
    julian = date2jd(int_year, int_month, int_day, 0.)
    return(julian)

###########################################################################
# Function to return the distance in arcsec between 2 RA-DEC coordinates
###########################################################################

def delta_theta(ra1, dec1, ra2, dec2):
    # all ra and dec in degrees

    inter = np.sin(np.radians(dec1))*np.sin(np.radians(dec2)) + \
            np.cos(np.radians(dec1))*np.cos(np.radians(dec2))*np.cos(np.radians(ra1 - ra2))
    if inter > 1.0:  # some rounding may result in this being >1.0000 which would throw an error
        inter = 1.0
    delta_theta_rad = np.arccos(inter)
    return(abs(3600.*np.degrees(delta_theta_rad)))



##########################################################################################
# Function to return a variable value given the corresponding keyword in
# the pest_pipeline_config.txt file
##########################################################################################

def get_conf_var(keyword, conf_list, default):  # keyword is string, conf_list is the list read from the config file
    expr = keyword + '\s*=\s*(.*)'      # will match 'keyword= variable' with or without spaces next to the =
    var_found = False
    for line in conf_list:
        try:
            var = re.search(expr, line).group(1)
            var_found = True
            break
        except:
            pass

    if not var_found:
        print(keyword + ' not found in pest_pipeline_config.txt.\nSetting to: ' + default)
        var = default
    return var

# end get_conf_var



##########################################################################################
# Function to convert 8.24 fixed to float
##########################################################################################

def fixed2float(val):                   # val is an integer
    if (val & (1 << (32 - 1))) != 0:    # if sign bit is set i.e. a negative number
        val = val - (1 << 32)           # compute the negative number

    out = float(val)/2**24  # division by 2**24 shifts the decimal point to the correct place

    return out

# end fixed2float()
###############################################################################################


###############################################################################################
# Function to extract data from a .phot file
# Returns: [frame_no, jd, frame_data, aperture_radius].
# where frame_no is the frame number derived from the numbering of the .phot file,
# jd is the Julian Date stamp from the frame,
# and frame_data is a nested list with the following structure;
# [[global_obj_id, x-coord, y_coord, sky, fwhm, mag, err], [global_obj_id2, x-coord, y_coord, sky, fwhm, mag, err],...]
#
# inputs are the filename, and aperture id
#
# Only matched objects are included.  Invalid values are replaced with NaN (numpy.nan)
#
##############################################################################################

def phot_data(phot_filename, ap_no):

    frame_data = []       # empty list to hold the result

    expr = r'tmp([0-9]+)\.pht$' # the filename must end with .pht
    sub_line = re.search(expr, phot_filename)
    frame_no = int(sub_line.group(1))
    f = open(phot_filename, 'rb')


    ##########################################################################################
    # read file header

    format_id = f.read(28).decode('utf-8')
    format_rev = int.from_bytes(f.read(4), byteorder='little')
    meta_len = int.from_bytes(f.read(4), byteorder='little')
    header_end = 36
    if format_id != 'C-Munipack photometry file\r\n':
        print('\n\nWarning:  This is not the photometry file expected!\n\n')
    if format_rev != 4:
        print('\nWarning: phot data format rev 4 expected. Actual: ', format_rev)

    print('Processing file: ', sub_line.group(), '= Frame', frm+1, 'of', tot_frames)


    ##########################################################################################
    # Metadata block

    dump = f.read(4)       # unused
    framex = int.from_bytes(f.read(4), byteorder='little')  # frame width in pixels
    framey = int.from_bytes(f.read(4), byteorder='little')  # frame height
    jdate = struct.unpack('<d', f.read(8))[0]      # little-endian binary64, access the first item in the tuple returned
    filter = f.read(70).decode('utf-8')

    f.seek(header_end + 280)
    matched = int.from_bytes(f.read(4), byteorder='little')  # 0 = not matched, 1 = matched

    f.seek(header_end + 320)
    obj_name = f.read(70).decode('utf-8')
    obj_ra = struct.unpack('<d', f.read(8))[0]
    obj_dec = struct.unpack('<d', f.read(8))[0]
    obs_loc = f.read(70).decode('utf-8')
    obs_long = struct.unpack('<d', f.read(8))[0]
    obs_lat = struct.unpack('<d', f.read(8))[0]
    xx = struct.unpack('<d', f.read(8))[0]
    xy = struct.unpack('<d', f.read(8))[0]
    xo = struct.unpack('<d', f.read(8))[0]

    meta_end = header_end + meta_len

    ############################################################################################
    # WCS data

    f.seek(meta_end)
    wcs_len = int.from_bytes(f.read(4), byteorder='little')


    wcs_end = meta_end + wcs_len + 4

    #############################################################################################
    # Table of apertures

    f.seek(wcs_end)
    num_ap = int.from_bytes(f.read(4), byteorder='little') # number of aperture records

    ap = []
    for idx in range(1, num_ap+1):
        ap_id = int.from_bytes(f.read(4), byteorder='little')
        ap_rad = struct.unpack('<d', f.read(8))[0]
        ap.append([ap_id, ap_rad])

    ap_end = wcs_end + num_ap*12 + 4

    #############################################################################################
    # Table of objects

    f.seek(ap_end)
    num_obj = int.from_bytes(f.read(4), byteorder='little') # number of object records

    obj_end = ap_end + num_obj*48 + 4

    matched_obj = 0                         # count the no of matched objects

    for idx in range(1, num_obj+1): # not sure if idx always == obj_id
        f.seek(ap_end + 4 + (idx - 1)*48)    # records start at offset 4 and have 48 bytes each
        obj_id = int.from_bytes(f.read(4), byteorder='little')
        global_obj_id = int.from_bytes(f.read(4), byteorder='little')
        x_coord = struct.unpack('<d', f.read(8))[0]
        y_coord = struct.unpack('<d', f.read(8))[0]
        sky = struct.unpack('<d', f.read(8))[0]
        sky_sd = struct.unpack('<d', f.read(8))[0]
        fwhm = struct.unpack('<d', f.read(8))[0]
        if fwhm > max_fwhm or fwhm < 0:
            fwhm = 0.0  # Muniwin sometimes returns FWHM=infinity, just set it to an invalid but benign number


    #############################################################################################
    # Table of measurements

    # default behavior of int.from_byte is 'signed = False',
    # so correct calc of negative nos is left to fixed2float()

        f.seek(obj_end + (idx-1)*12*num_ap + (ap_no-1)*12)  # there seems to be 12 bytes a record, not 8

        instr_mag = fixed2float(int.from_bytes(f.read(4), byteorder='little'))
        mag_err = fixed2float(int.from_bytes(f.read(4), byteorder='little'))

        #########################################################################
        # Eliminate the unmatched objects, and replace invalid values with NaN

        if global_obj_id != 0xFFFFFFFF:     # object is matched
            matched_obj += 1
            if instr_mag > 100:     # the mag is invalid
                instr_mag = np.nan
                mag_err = np.nan

            datalist = [global_obj_id, x_coord, y_coord, sky, fwhm, instr_mag, mag_err ]
            frame_data.append(datalist)



    f.close()
    print('Matched: ', matched_obj, ' of ', num_obj)
    return [frame_no, jdate, frame_data, ap[ap_no-1][1], filter, matched_obj]

# End phot_data()
####################################################################################################

####################################################################################################
# Function to plot a light curve wth error bars
####################################################################################################
def plot_lc_err(bjd, mag, err, meanmag, std, title, filename):
    # title, filename are strings, meanmag, std: float, the rest arrays

    plt.figure(None, figsize=(12,6))
    plt.errorbar(bjd - 2450000, mag, yerr=err, fmt='bo', lw=1, capsize=2)
    ax = plt.gca()
    ax.get_xaxis().get_major_formatter().set_useOffset(False) # turn off scientific notation
    ax.invert_yaxis()

    plt.rc('grid', linestyle='dotted', color='gray')
    plt.title(title)
    if date_type == 'HJD':
        plt.xlabel('HJD (UTC) - 2450000')
    else:
        plt.xlabel('BJD (TDB) - 2450000')
    plt.ylabel('Mag')
    plt.text(0.98, 0.95, 'Mean = ' + str(round(meanmag,3)), \
             horizontalalignment='right', transform=ax.transAxes)
    plt.text(0.98, 0.9, 'SD = ' + str(round(std,4)), \
             horizontalalignment='right', transform=ax.transAxes)
    plt.text(0.98, 0.85, 'Points = ' + str(np.count_nonzero(~np.isnan(mag))), \
             horizontalalignment='right', transform=ax.transAxes)
    plt.text(0.02, 0.95, 'Target: ' + target_name, \
             horizontalalignment='left', transform=ax.transAxes)
    plt.text(0.02, 0.9, 'Filter: ' + filter, \
             horizontalalignment='left', transform=ax.transAxes)
    plt.text(0.02, 0.85, 'Obs Date: ' + init_time.iso, \
             horizontalalignment='left', transform=ax.transAxes)
    plt.text(0.0, -0.1, observatory + ' Observatory', color='gray', \
         horizontalalignment='left', transform=ax.transAxes)

    plt.grid(True)
    plt.savefig(filename,dpi=72)
    plt.close("all")        # clears memory

# End plot_lc_err()
###################################################################################################

####################################################################################################
# Function to plot a light curve with small dots and transit box
####################################################################################################
def plot_lc(bjd, mag, err, meanmag, std, title, filename, ingress, egress, NEBmag, NEBppt, disp):
    # title, filename, ingress, egress, disp are strings: meanmag, std, NEBmag, NEBppt: float, the rest arrays

    fingress = float(ingress)
    fegress = float(egress)

    fig, ax1 = plt.subplots(figsize=(12,6))
    plt.rc('grid', linestyle='dotted', color='gray')

    ax1.plot(bjd - 2450000, mag-meanmag,'b.')

    ax1.get_xaxis().get_major_formatter().set_useOffset(False) # turn off scientific notation
    ax1.get_yaxis().get_major_formatter().set_useOffset(False) # turn off scientific notation

    ax1.invert_yaxis()
    ax1.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.3f'))


    # allow some margin around the data
    ax1.margins(0.05)


    start, end = ax1.get_xlim()  # get the start and end limits of the x-axis
    low, high = ax1.get_ylim()

    ax1.set_ylim(low, high)  # for some reason, get does not return the correct values. To align left and right
    # y axes this had to be done.

    tick_spacing = math.floor((end-start)*10.0)/100  # makes sure there are about 10 ticks on the x-axis
    if tick_spacing > 0:
        ax1.xaxis.set_major_locator(ticker.MultipleLocator(tick_spacing))

    #####################################################
    # Draw a box representation of the required NEB transit to cause the detected signal

    if NEBmag == -1:
        no_bottom = True
    else:
        no_bottom = False


    if ingress != '0' and start > fingress:
        start = fingress  # avoid the box transit doubling back on itself
    if egress != '0' and end < fegress:
        end = fegress

    if ingress == '0' and egress == '0':
        # plot a horizontal at the mean of the data if no transit timing given
        plt.axhline(y=0.0, color='gray', linestyle='--', lw=2)
    elif ingress != '0':
        plt.hlines(0.0,start,fingress, color='maroon', linestyle='-', lw=2)
        if no_bottom:
            plt.vlines(fingress,0.0,low, color='maroon', linestyle='-', lw=2)
        else:
            plt.vlines(fingress,0.0,NEBmag, color='maroon', linestyle='-', lw=2)

        if egress != '0':
            if no_bottom:
                plt.vlines(fegress,0.0,low, color='maroon', linestyle='-', lw=2)
            else:
                plt.hlines(NEBmag,fingress,fegress, color='maroon', linestyle='-', lw=2)
                plt.vlines(fegress,0.0,NEBmag, color='maroon', linestyle='-', lw=2)
            plt.hlines(0.0,fegress,end, color='maroon', linestyle='-', lw=2)

        elif not no_bottom:
            plt.hlines(NEBmag,fingress,end, color='maroon', linestyle='-', lw=2)
    else:
        if no_bottom:
            plt.vlines(fegress,0.0,low, color='maroon', linestyle='-', lw=2)
        else:
            plt.hlines(NEBmag,start,fegress, color='maroon', linestyle='-', lw=2)
            plt.vlines(fegress,0.0,NEBmag, color='maroon', linestyle='-', lw=2)
        plt.hlines(0.0,fegress,end, color='maroon', linestyle='-', lw=2)


    # find where along the y-axis to put the ingress and egress labels
    y_pos = low

    # plot vertical lines to mark ingress and egress
    # note the adjustments to the text coords are due to us putting a margin around the data
    if ingress != '0':
        plt.axvline(x=float(ingress), color='green', linestyle='--', lw=1, ymin=0.1)
        ax1.annotate('Pred\nIngress', xy=(fingress, y_pos), xytext=(fingress, y_pos),
                     horizontalalignment='center', verticalalignment='bottom', color='green')

    if egress != '0':
        plt.axvline(x=fegress, color='red', linestyle='--', lw=1, ymin=0.1)
        ax1.annotate('Pred\nEgress', xy=(fegress, y_pos), xytext=(fegress, y_pos),
                     horizontalalignment='center', verticalalignment='bottom', color='red')


    plt.title(title)
    if date_type == 'HJD':
        ax1.set_xlabel('HJD (UTC) - 2450000')
    else:
        ax1.set_xlabel('BJD (TDB) - 2450000')
    ax1.set_ylabel('delta-mag')
    ax1.figure.autofmt_xdate()  # to avoid overlapping x labels


    # make the Mag axis
    ax2 = ax1.twinx()
    ax2.set_ylim(low+meanmag, high+meanmag)
    ax2.set_ylabel('Mag')
    ax1.grid(True)
    ax2.grid(False)
    ax2.yaxis.set_major_formatter(ticker.FormatStrFormatter('%.3f'))



    #plt.text(0.98, 0.95, 'Mean = ' + str(round(meanmag,3)), \
    #         horizontalalignment='right', transform=ax1.transAxes)
    plt.text(0.98, 0.95, 'SD = ' + str(round(std,4)), \
             horizontalalignment='right', transform=ax1.transAxes)
    plt.text(0.98, 0.9, 'NEBdepth = '+str(round(NEBppt)) + 'ppt; '+str(round(NEBmag*1000))+'mmag', \
             horizontalalignment='right', transform=ax1.transAxes)
    plt.text(0.98, 0.85, 'NEB disposition: ' + disp, \
             horizontalalignment='right', transform=ax1.transAxes)
    plt.text(0.02, 0.95, 'Target: ' + target_name + ' in ' + filter, \
             horizontalalignment='left', transform=ax1.transAxes)

    plt.text(0.02, 0.9, 'Obs Date: ' + obs_date, \
             horizontalalignment='left', transform=ax1.transAxes)
    plt.text(0.0, -0.18, observatory + ' Observatory', color='gray', \
             horizontalalignment='left', transform=ax1.transAxes)

    plt.tight_layout()
    plt.savefig(filename,dpi=72)
    plt.close("all")        # clears memory

# End plot_lc()
###################################################################################################




###################################################################################################
#
#
# Main part of the program
#
#
###################################################################################################


# read config file

homedir = os.path.expanduser('~')       # find the home directory

config = open(os.path.join(homedir,'bin','pest_pipeline_config.txt'), 'r').readlines()
config = [x.strip() for x in config]  # remove leading and trailing whitespaces

for i in range(len(config)):
    if config[i] != '':
        if config[i][0] == '#':
            config[i] = ''      # set all comment lines to blank

param_path = get_conf_var('magpy_param_path', config, 'NA')
catalog_dir = get_conf_var('catalog_dir', config, 'NA')
proj_dir = get_conf_var('proj_dir', config, 'NA')
observatory = get_conf_var('observatory', config, 'NA')
telescope = get_conf_var('telescope', config, 'NA')
camera = get_conf_var('camera', config, 'NA')
tess_user = get_conf_var('tess_user', config, 'NA')
date_type = get_conf_var('date_type', config, 'BJD')

scale = float(get_conf_var('scale', config, 'NA'))   # arcsec per pixel
add_stars = int(get_conf_var('add_stars', config, '25'))   # no of additional stars to be auto-selected
star_limit = int(get_conf_var('star_limit', config, '35'))   # limit to total no of selected stars
min_stars = int(get_conf_var('min_stars', config, '20'))   # stars are added if current selection is fewer than this
outlier_th = float(get_conf_var('outlier_threshold', config, '1.0'))   # smaller number = more points become outliers
proximate_dist = float(get_conf_var('proximate_dist', config, '2.5'))   # NEB clearance dist in min
plot_dist = float(get_conf_var('plot_dist', config, '2.5'))   # zoomed finder plot and dmag/SD plot dist limit in min

#catalog_path = get_conf_var('catalog_dir', config, 'NA')
plot_north = get_conf_var('plot_north', config, 'Yes')

if plot_north == 'No' or plot_north == 'no':
    rotate = False
else:
    rotate = True

dmag_buf = float(get_conf_var('dmag_buf', config, '0.5'))  # subtract from dmag for NEBdepth calc
gaia = get_conf_var('check_gaia', config, 'Yes')

if gaia == 'No' or gaia == 'no':
    check_gaia = False
else:
    check_gaia = True

# max_theta in arcsec. If no detected star is found within this dist of a Gaia star, consider it not detected
max_theta = float(get_conf_var('max_theta', config, '3.0'))

# If the detected star has mag more that this different from Gaia g_mag, consider it not detected
max_delta_mag = float(get_conf_var('max_delta_mag', config, '2.0'))

# plot_dist will override if it is less than proximate_dist
if plot_dist < proximate_dist:
    proximate_dist = plot_dist

scale_low = str(scale - 0.2)
scale_high = str(scale + 0.2)                # used to tell plate solver bounds for the plate-scale

# detrend switches

detrend = True
detrend_rad = True
detrend_fwhm = True
detrend_airmass = True
detrend_sky = True
detrend_plots = True

detrend_param_list = []

detrend_sw = get_conf_var('detrend', config, 'Yes')
if detrend_sw == 'No' or detrend_sw == 'no':
    detrend = False

detrend_sw = get_conf_var('detrend_rad', config, 'Yes')
if detrend_sw == 'No' or detrend_sw == 'no':
    detrend_rad = False
else:
    detrend_param_list.append('Radial-dist')

detrend_sw = get_conf_var('detrend_fwhm', config, 'Yes')
if detrend_sw == 'No' or detrend_sw == 'no':
    detrend_fwhm = False
else:
    detrend_param_list.append('FWHM')

detrend_sw = get_conf_var('detrend_airmass', config, 'Yes')
if detrend_sw == 'No' or detrend_sw == 'no':
    detrend_airmass = False
else:
    detrend_param_list.append('Airmass')

detrend_sw = get_conf_var('detrend_sky', config, 'Yes')
if detrend_sw == 'No' or detrend_sw == 'no':
    detrend_sky = False
else:
    detrend_param_list.append('Sky')

detrend_sw = get_conf_var('detrend_plots', config, 'Yes')
if detrend_sw == 'No' or detrend_sw == 'no':
    detrend_plots = False

solve = get_conf_var('plate_solve', config, 'Yes')
if solve == 'No' or solve == 'no':
    plate_solve = False
    check_gaia = False
else:
    plate_solve = True


# this one option will switch off outlier LC plots, detrending, plate-solving, and Gaia check
simple = get_conf_var('simple_mode', config, 'No')
if simple == 'Yes' or simple == 'yes':
    simple_mode = True
    outlier_plots = False
    detrend = False
    plate_solve = False
    check_gaia = False
else:
    simple_mode = False
    outlier_plots = True


#########################################################################################



# define directory names

plots_dir = os.path.relpath('detrend_plots', start=os.curdir)

f_pcinfo = open('PESTphot_info.txt', 'w')
f_pcinfo.write('PESTphot ' + ver_no + '\n')
f_pcinfo.write('Ensemble star detrending: ')
for detrend_param in detrend_param_list:
    f_pcinfo.write(detrend_param + ' ')

##################################################################################################
#
# Sets the in_path to the current directory, and obtains obj_id and obs_date from the path
#
##################################################################################################

in_path = os.getcwd()

target_name = os.path.basename(os.path.dirname(in_path))
obs_date = os.path.basename(in_path)
obs_date_strip = re.sub('[-.]', '', obs_date)  # strips out any dash or dot in the directory name

print('Target ID: ', target_name, '\n')


# exit if the catalog file does not exist
if not os.path.exists(catalog_dir + target_name + ".xml"):
    print("\nError: No catalog file found for target " + target_name)
    sys.exit()





#################################################################################################


# set limit to how high FWHM can be. Needed because in some (invalid) situations Muniwin returns a large number
max_fwhm = 100


# check which aperture number have been used in previous runs, if any
old_ap = None
old_ap_list = []
ap_inconsistent = False
muni_file_list = glob.glob(os.path.join(os.pardir,'*','Muni*.txt'))
if len(muni_file_list) > 0:
    for muni_filename in muni_file_list:
        fm_lines = open(muni_filename).readlines()
        if re.search(r"Aperture:\s+([0-9]+),", fm_lines[1]):  # this line has the aperture no
            sub_line = re.search(r"Aperture:\s+([0-9]+),", fm_lines[1])
            this_ap = sub_line.group(1)
        old_ap_list.append(this_ap)
        if old_ap != None and old_ap != this_ap:
            ap_inconsistent = True
        else:
            old_ap = this_ap

    if ap_inconsistent:
        print('WARNING: Previous observations used inconsistent aperture nos: ', end = '', flush = True)
        print(old_ap_list)
    else:
        print('Previous observations used aperture no: ' + old_ap)
        print('Press Enter when input requested to use this.')


sel_ap = None      # if not changed by a proper passed parameter below, will trigger an error

#date_type = 'BJD'  # this is the default, unless changed to HJD

if len(sys.argv) not in range(1,3):
    sys.exit('ERROR: PESTphot expects either 0 or 1 parameter.\n')
elif len(sys.argv) == 1:  # no command line parameter
    if old_ap == None:
        sel_ap_str = input('Input aperture number between 1 and 12 incl: '.strip())
    else:
        sel_ap_str = input('Input aperture number between 1 and 12 incl: ({:s}) '.format(old_ap)).strip()

    if old_ap != None and sel_ap_str == '':  # ie the user has just pressed Enter
        sel_ap_str = old_ap
else:  # command line parameter given, which should be the aperture number
    if old_ap == None:
        sel_ap_str = sys.argv[1]
    elif sys.argv[1] != old_ap and not ap_inconsistent:
        print('ERROR: You are using an inconsistent aperture number!')
        sys.exit()
    elif sys.argv[1] != old_ap and ap_inconsistent:
        print('WARNING: Your previous and current aperture numbers are inconsistent!')
        sel_ap_str = sys.argv[1]
    else:
        sel_ap_str = sys.argv[1]



try:
    if int(sel_ap_str) in range(1,13):
        sel_ap = int(sel_ap_str)
    else:
        sys.exit('ERROR: aperture number must be between 1 and 12 inclusive.\n')

except ValueError:
    sys.exit('ERROR: no valid aperture number entered.\n')


################################


# find active projects and if necessary ask the user to choose between them

active_proj_paths = []
for root, dirs, files in os.walk(proj_dir):
    if any('.pht' in s for s in files):
        active_proj_paths.append(root)
if len(active_proj_paths) == 0:
    print('ERROR: No active project found.\n')
elif len(active_proj_paths) > 1:
    print('More than one active project found.')
    for idx, proj in enumerate(active_proj_paths, 1):
        print(idx, proj)

    sel_proj_dir = active_proj_paths[int(input('Type the number to select: '))-1]
else:
    sel_proj_dir = active_proj_paths[0]

print('Working on the project at:', sel_proj_dir)
print('')

#########################################################################################
# Copy over latest INFO.txt file found under the same target directory
# and give user the choice of whether to overwrite the MagPy params file

info_file_list = glob.glob(os.path.join(os.pardir,'*','INFO.txt'))
if len(info_file_list) > 0:
    latest_file = max(info_file_list, key=os.path.getctime)
    print('Found previous MagPy parameters file for this target: ' + latest_file)
    in_lines = open(latest_file).readlines()
    info_obj_id = in_lines[11].strip()

    if not os.path.exists('INFO.txt'):
        copyfile(latest_file, os.path.join(in_path, 'INFO.txt'))

    if os.path.exists(os.path.join(param_path, 'param_MagPy.txt')):  # find out what target is in the MagPy file
        in_lines = open(os.path.join(param_path, 'param_MagPy.txt')).readlines()
        file_obj_id = in_lines[11].strip()
        if file_obj_id == info_obj_id:
            print('Target in existing INFO.txt is consistent with MagPy parameters file: ' + info_obj_id)
        else:
            print('Current MagPy parameters file has a different target: ' + file_obj_id)
    overwrite_param = input('Overwrite MagPy parameters? (Y/y for Yes, Enter or anything else for No) ').strip()
    if overwrite_param == 'Y' or overwrite_param == 'y':
        copyfile(latest_file, os.path.join(param_path, 'param_MagPy.txt'))


##########################################################################################
# Get other inputs
##########################################################################################

var_count = 0       # will set to 1 when there is a star tagged 'var' in the catalog
comp_count = 0      # ditto for the 'comp' star

mt_lines = ['' for x in range(3)]

# use the INFO.txt file for COMP if there hasn't been a previous run
if os.path.exists('INFO.txt') and not os.path.exists('PESTphot_inputs.txt'):
    print('\nReading from INFO.txt. Press Enter at the prompts to use those values.')
    in_lines = open('INFO.txt').readlines()
    mt_lines[0] = in_lines[5].rstrip()
    print('COMP mag: ' + mt_lines[0].strip())


if os.path.exists('PESTphot_inputs.txt'):
    print('\nFor current default values (in brackets) press Enter at input prompts.')
    mt_lines = open('PESTphot_inputs.txt').readlines()
    #print('COMP mag: ' + mt_lines[0].strip())
    #print('Ingress: ' + mt_lines[1].strip())
    #print('Egress: ' + mt_lines[2].strip())

# format how the default values are presented at the input prompt
def_val = []
for i in range(3):
    if mt_lines[i].strip() == '':
        if i == 0:
            def_val.append('')
        else:
            def_val.append('(0)')
    else:
        def_val.append('({:s})'.format(mt_lines[i].strip()))

ref_cm_str = input('Enter magnitude of COMP: {:s} '.format(def_val[0])).strip()
if ref_cm_str == '':
    ref_cm = float(mt_lines[0])
else:
    ref_cm = float(ref_cm_str)

ingress = input('HJD/BJD-2450000 of Ingress (0 if not needed): {:s} '.format(def_val[1])).strip()
if ingress == '':
    ingress = mt_lines[1].strip()
egress = input('HJD/BJD-2450000 of Egress (0 if not needed): {:s} '.format(def_val[2])).strip()
if egress == '':
    egress = mt_lines[2].strip()

# catch the case where the user presses Enter but there are no existing default values
if ingress == '':
    ingress = '0'
if egress == '':
    egress = '0'

# write this info for use by magpy
f_mt = open('PESTphot_inputs.txt', 'w')
f_mt.write(str(ref_cm) + '\n')
f_mt.write(ingress + '\n')
f_mt.write(egress + '\n')
f_mt.close()



# get the TESS observed transit depth for NEB clearance calcs
td_lines = ['']
if os.path.exists('td.txt'):  # records the depth for use in future runs
    td_lines = open('td.txt').readlines()
    print('\nCurrent default TESS depth = ' + td_lines[0].strip())
    print('To keep this value press Enter when input is requested.')

if td_lines[0].strip() == '':
    def_depth_str = '(0)'
else:
    def_depth_str = '({:s})'.format(td_lines[0].strip())


TESSdepth_input = input('Exp transit depth (append "mmag" or "ppt"; '
                        'Enter 0 to turn off Gaia check): {:s} '.format(def_depth_str).strip())
if TESSdepth_input == '':
    TESSdepth_input = td_lines[0].strip()  # accept the default value

f_td = open('td.txt', 'w')
f_td.write(TESSdepth_input + '\n')
f_td.close()

if re.search(r"([.0-9]+)\s*([a-z]*)", TESSdepth_input):
    sub_line = re.search(r"([.0-9]+)\s*([a-zA-Z]*)", TESSdepth_input)
    depth_input = float(sub_line.group(1))
    if depth_input > 0.0:
        if sub_line.group(2) == 'mmag' or sub_line.group(2) == 'MMAG':
            TESSdepth = 1000*(1-10**(-depth_input/1000/2.5))
            TESSdepth_mmag = depth_input
            print('Converted '+str(depth_input)+'mmag to '+str(round(TESSdepth,2))+'ppt.\n')
        else:
            TESSdepth = depth_input
            TESSdepth_mmag = -2.5*math.log10(1.0-TESSdepth/1000)*1000.0
            print('Using '+str(depth_input)+'ppt or '+str(round(TESSdepth_mmag,2))+'mmag.\n')
    else:
        print('Either 0 or -ve depth entered. Setting depth to 0 and turning Gaia check off.\n')
        check_gaia = FALSE
        TESSdepth = 0.0
        TESSdepth_mmag = 0.0
else:
    print('No depth entered. Setting depth to 0 and turning Gaia check off.\n')
    check_gaia = FALSE
    TESSdepth = 0.0
    TESSdepth_mmag = 0.0



###################################################################################################
#
#
# Main part of the program
#
#
###################################################################################################

if not os.path.exists(catalog_dir + target_name + '.xml'):
    print('\n **** ERROR.  Catalog file not present *** ' + catalog_dir + target_name + '.xml\n\n')
    sys.exit()


# make a subdirectory for the plots
if not os.path.exists(plots_dir) and detrend_plots and detrend:
    os.mkdir(plots_dir)


####################################################################################
# Parse the reference xml file

xmldoc = minidom.parse(catalog_dir + target_name + '.xml')
print('Parsing reference XML file...')

#refjd_el = xmldoc.getElementsByTagName('jd')
#refjd = refjd_el[0].childNodes[0].nodeValue

try:
    exptime_el = xmldoc.getElementsByTagName('exptime')
    exptime = exptime_el[0].childNodes[0].nodeValue
except:
    exptime = input('XML file has no exp time. Input(s): ').strip()

#filter_el = xmldoc.getElementsByTagName('filter')
#filter_ref = filter_el[0].childNodes[0].nodeValue  # the filter in the xml file may be different from the obs

#object_el = xmldoc.getElementsByTagName('object')
#object = object_el[0].childNodes[0].nodeValue

try:
    ra2000_el = xmldoc.getElementsByTagName('ra2000')
    ra2000 = ra2000_el[0].childNodes[0].nodeValue

    dec2000_el = xmldoc.getElementsByTagName('dec2000')
    dec2000 = dec2000_el[0].childNodes[0].nodeValue

    longitude_el = xmldoc.getElementsByTagName('longitude')
    longitude = longitude_el[0].childNodes[0].nodeValue

    latitude_el = xmldoc.getElementsByTagName('latitude')
    latitude = latitude_el[0].childNodes[0].nodeValue
except:
    sys.exit('ERROR: target coords and/or observatory location not in catalog XML.\n')

select_list = xmldoc.getElementsByTagName('select')
print('Number of selected stars: ', len(select_list))
selected_stars = []     # will hold tuples of (id, label)
for s in select_list:
    selected_stars.append((s.attributes['id'].value, s.attributes['label'].value))
    if s.attributes['label'].value == 'var':
        var_star_id = s.attributes['id'].value

    elif s.attributes['label'].value == 'comp':
        comp_star_id = s.attributes['id'].value

# make the list of selected stars into a dataframe with the star labels as index
sel_df = pd.DataFrame(selected_stars, columns=['id', 'label'])
sel_df.set_index('label', inplace=True)

star_list = xmldoc.getElementsByTagName('s')
print('Number of stars: ', len(star_list))
stars = []     # will hold tuples of (id, x, y)
proximate_stars = []  # will hold ids of stars within proximate_dist of VAR

for s in star_list:
    stars.append((s.attributes['id'].value, s.attributes['x'].value, s.attributes['y'].value))
    if s.attributes['id'].value == var_star_id:
        var_x = float(s.attributes['x'].value)
        var_y = float(s.attributes['y'].value)

# now identify stars closer than proximate_dist
for s in star_list:
    star_x = float(s.attributes['x'].value)
    star_y = float(s.attributes['y'].value)
    if scale * math.sqrt((star_x - var_x)**2 + (star_y - var_y)**2) <= proximate_dist*60. and s.attributes['id'].value != var_star_id:
        proximate_stars.append(s.attributes['id'].value)


# make the list of stars into a dataframe with id as index
star_df = pd.DataFrame(stars, columns=['id', 'x', 'y'])
star_df.set_index('id', inplace=True)

all_stars = star_df.index.values        # an array with ids of all matched stars

# get ccd size
frame_size = xmldoc.getElementsByTagName('stars')
framex = frame_size[0].attributes['width'].value
framey = frame_size[0].attributes['height'].value
print('CCD dimensions: ', framex, framey)



# end of XML parsing
###############################################################################################

# Define objects that describe target and observatory coordinates
try:
    target_coord = coord.SkyCoord(ra2000, dec2000, unit=(u.hourangle, u.deg), frame='icrs')
    print('Target ', target_name, ' is at coords ', target_coord.to_string(style='hmsdms'))

    # convert E/W and N/S forms of long and lat to positive or negative
    if longitude[-1] == 'E':
        s_longitude = longitude[:-1]
    elif longitude[-1] == 'W':
        s_longitude = '-'+longitude[:-1]
    else:
        print('Error: Either E or W expected at end of target coordinates.')
        raise Exception()

    long_obj = coord.Longitude(s_longitude, unit=u.deg)

    if latitude[-1] == 'N':
        s_latitude = latitude[:-1]
    elif latitude[-1] == 'S':
        s_latitude = '-'+latitude[:-1]
    else:
        print('Error: Either N or S expected at end of observatory coordinates.')
        raise Exception()

    lat_obj = coord.Latitude(s_latitude, unit=u.deg)
    print('Observatory coords: ',long_obj, lat_obj)

    obs_location = coord.EarthLocation.from_geodetic(long_obj, lat_obj)
except:
    sys.exit('ERROR: target coords and/or observatory location in catalog XML not valid.\n')

###############################################################################################
# Summary of data structures:
#
# stars:  type=list of tuples; tuples are (id, x, y) of all matched stars; element type=str
# star_df: type=dataframe; index=id,  columns=x, y for all matched stars; element type=str
# all_stars: type=array: id of all matched stars; ; element type=str
# star_timeseries: type=list of (list of lists); timeseries for all matched stars;
#   eg. star_timeseries[0]: list of lists with elements [frame_no,jd,x,y,sky,fwhm,mag,err] ie
#   a single timestep. In total this is the full timeseries for each star.
# df2_list: list of dataframes. Each dataframe contains the full timeseries for a star. Order of
#   dataframes is the same as in the all_stars array
# sel_df: type=dataframe: index=label, column=id for all selected stars
# vc_stat_df: type=dataframe: index=id, mean, sd, good-pts


###############################################################################################
# Read photometry files and structure the data

# get full list of .pht files
phot_file_list = glob.glob(os.path.join(sel_proj_dir, '*.pht'))
if phot_file_list == []:
    print('Error: No .pht files found...\n')

phot_file_list.sort()
tot_frames = len(phot_file_list)


headings = ['obj_id', 'x_coord', 'y_coord', 'sky', 'fwhm', 'mag', 'err']
headings2 = ['frame_no', 'jd', 'bjd', 'airmass', 'x_coord', 'y_coord', 'sky', 'fwhm', 'mag', 'err']
frm = 0        # initialise the frame count.


star_timeseries = []  # initialise the list that will hold timeseries data for all matched stars
for idx in range(len(all_stars)):
    star_timeseries.append([])  # we end up with the same number of empty lists as matched stars

#best_fwhm = 100  # variable to track best median fwhm amongst the frames. Start with arbitrary large number
best_matched = 0  # variable to track max no of matched objects in a frame
best_frame = ''  # variable to hold the filename (.pht) of the best frame found so far

for phot_filename in phot_file_list:
    frame_results = phot_data(phot_filename, sel_ap)

    frame_no = frame_results[0]
    jd = frame_results[1]
    df = pd.DataFrame(frame_results[2], columns=headings)
    df.set_index('obj_id', inplace=True)

    matched = frame_results[5]

    #frame_fwhm = np.nanmedian(df['fwhm'].values)  # calc median fwhm of all stars in the field
    if matched > best_matched:  # find the frame with max matched objects
        best_matched = matched
        best_frame = phot_filename
        best_frame_no = frame_no
        #print(str(best_matched) + ': in best frame: ' + best_frame)

    # calculate light travel time and convert JD(UTC) either to BJD(TDB), or HJD
    times = time.Time(jd, format='jd', scale='utc', location=obs_location)
    if date_type == 'HJD':
        ltt_helio = times.light_travel_time(target_coord, 'heliocentric')
        bjd_time = times.utc + ltt_helio
    else:
        ltt_bary = times.light_travel_time(target_coord)
        bjd_time = times.tdb + ltt_bary

    bjd = bjd_time.jd  # note here that although the variable name is bjd, the format may be HJD

    # calculate airmass
    target_altaz = target_coord.transform_to(coord.AltAz(obstime=times, location=obs_location))
    airmass = target_altaz.secz.value       # secz returns a Quantity object - .value accesses the float

    # note the timestamp of the first frame
    if frm == 0:
        init_jd = jd
        init_time = time.Time(init_jd, format='jd', out_subfmt='date', scale='utc')
        filter_raw = frame_results[4].strip()       # the filter is defined by the first frame
        if filter_raw == 'Vc':               # catched a couple of quirks in filter naming used by PEST
            filter = 'V'
        elif filter_raw == 'Bc':
            filter = 'B'
        else:
            filter = filter_raw

    idx = 0
    for star_no in all_stars:
        timestep = [frame_no, jd, bjd, airmass]
        try:
            timestep_data = df.loc[int(star_no)]
            timestep.extend(timestep_data.tolist())
            # results in a list with full timestep data for each star ie
            # [frame_no,jd,bjd,airmass,x,y,sky,fwhm,mag,err]

            # determine the sky level around target on the best frame
            if star_no == var_star_id and frame_no == best_frame_no:
                best_frame_sky = timestep_data.tolist()[2]
                #print('/nBest frame sky = ' + str(best_frame_sky))

        except KeyError:    # if this star is not found in this frame, fill the timestep w NaN
            timestep.extend([np.nan,np.nan,np.nan,np.nan,np.nan,np.nan])
            pass

        star_timeseries[idx].append(timestep)

        idx += 1
    frm += 1
final_jd = jd
obs_dur = (final_jd - init_jd)*24*60   # duration in mins

# copy the FITS file from the best frame to here
best_frame = best_frame.replace(".pht", ".fts")

copyfile(best_frame, os.path.join(in_path, 'best_frame.fts'))


print('No. of Frames: ', frm)
print('Aperture No: ', sel_ap, ' with radius ', frame_results[3])
print('Building timeseries for all', len(star_list),'detected stars.')
f_pcinfo.write('\nAperture No: ' + str(sel_ap) + ' with radius ' + str(frame_results[3]))
f_pcinfo.write('\nObservation Duration (mins): ' + str(round(obs_dur,1)))
f_pcinfo.write('\nNo. of frames: ' + str(frm))
f_pcinfo.write('\nFilter: ' + filter)

df2_list = []
# df2_list will contain timeseries dataframes for each matched star in the same order
# as in the all_stars array

for idx in range(len(all_stars)):
    df2_list.append(pd.DataFrame(star_timeseries[idx], columns=headings2))
    df2_list[idx].set_index('frame_no', inplace=True)
    df2_list[idx].sort_index(inplace=True)
    print('.', end='', flush=True)


# store some of these parameters for use for the ebcheck part of the script
phot_rad = float(frame_results[3])
phot_rad_deg = phot_rad * scale/60/60


# some definitions, now that we have read the data

std_name = target_name + '_' + obs_date_strip + '_' + observatory + '_' + filter  # the standard part of key output filenames

out_filename_template = "_ebcheck.txt"      # holds data to plot stars
hood_filename = std_name + "_neighbourhood.txt"         # id, RA, DEC, mag, scatter for neighbour stars
magdev_filename = "magdev.txt"      # file containing all detected stars, their delta mag to comp and scatter



# All data has now been read and structured
###################################################################################################
#
#  Plotting and de-trending
#
###################################################################################################


print('\nMaking plots...')


##################################################################################################
# Plot graphs on one sheet
##################################################################################################

fig_com, (ax10, ax20, ax30) = plt.subplots(3, 1, sharex=True, figsize=(12,18))


# plot airmass and FWHM for the target

idxv = np.where(all_stars == str(var_star_id))[0][0]  # locating the target star in the all_stars array
idxc = np.where(all_stars == str(comp_star_id))[0][0]  # locating the COMP star

bjd_a = df2_list[idxv]['bjd'].values     # for the target
bjdc_a = df2_list[idxc]['bjd'].values   # for COMP

airmass_a = df2_list[idxv]['airmass'].values
fwhm_a = df2_list[idxc]['fwhm'].values  # use FWHM for COMP because VAR is sometimes intentionally saturated
fwhm_mean = np.nanmean(fwhm_a)


plt.rc('grid', linestyle='dotted', color='gray')
ax10.plot(bjd_a - 2450000, airmass_a, 'bo')

# Make the y-axis label, ticks and tick labels match the line color.
ax10.set_ylabel('Airmass', color='b')
ax10.tick_params('y', colors='b')
ax10.grid(axis='x', linestyle='dotted', color='gray')
ax10.invert_yaxis()

ax11 = ax10.twinx()
ax11.plot(bjd_a - 2450000, fwhm_a, 'r.')
ax11.set_ylabel('FWHM (pix)', color='r')
ax11.tick_params('y', colors='r')

ax10.set_title('Airmass and FWHM')


ax10.get_xaxis().get_major_formatter().set_useOffset(False)  # turn off scientific notation
ax11.grid(True)

plt.text(0.02, 0.95, 'Target: ' + target_name, \
         horizontalalignment='left', transform=ax11.transAxes)
plt.text(0.02, 0.9, 'Filter: ' + filter, \
         horizontalalignment='left', transform=ax11.transAxes)
plt.text(0.02, 0.85, 'Obs Date: ' + init_time.iso, \
         horizontalalignment='left', transform=ax11.transAxes)


f_pcinfo.write('\nFWHM mean: ' + str(round(fwhm_mean,2)) + ' pix; ' + str(round(fwhm_mean*scale,2)) + ' arcsec')
f_pcinfo.write('\nBest Frame: ' + best_frame)
f_pcinfo.write('\nExp Time: ' + exptime)


###########################################################################################
# plot X and Y shifts for the target

x_a = df2_list[idxv]['x_coord'].values
y_a = df2_list[idxv]['y_coord'].values


ax20.plot(bjd_a - 2450000, x_a, 'b.')

ax20.set_ylabel('X position', color='b')
ax20.tick_params('y', colors='b')

ax20.grid(axis='x')

ax21 = ax20.twinx()
ax21.plot(bjd_a - 2450000, y_a, 'r.')
ax21.set_ylabel('Y position', color='r')
ax21.tick_params('y', colors='r')

ax20.set_title('XY position of target')

ax21.get_xaxis().get_major_formatter().set_useOffset(False)  # turn off scientific notation
ax21.grid(True)

###########################################################################################
# plot raw instr mags for the target and COMP

var_mag_a = df2_list[idxv]['mag'].values
var_err_a = df2_list[idxv]['err'].values

comp_mag_a = df2_list[idxc]['mag'].values
comp_err_a = df2_list[idxc]['err'].values

var_mean = np.nanmean(var_mag_a)
comp_mean = np.nanmean(comp_mag_a)
offset = 0.1    # an arbitrary offset for the two datasets



ax30.errorbar(bjd_a - 2450000, var_mag_a - var_mean, yerr=var_err_a, fmt='bo', label='Target',capsize=2)
ax30.errorbar(bjdc_a - 2450000, comp_mag_a - comp_mean + offset, yerr=comp_err_a, fmt='ro', label='COMP',capsize=2)

ax30.set_ylabel('Instr mag (arbitrary offset)')
ax30.set_title('Instr Mag; Target and COMP')


ax30.invert_yaxis()

ax30.get_xaxis().get_major_formatter().set_useOffset(False)  # turn off scientific notation
ax30.grid(True)
ax30.legend(loc='lower right')

plt.text(0.0, -0.1, observatory + ' Observatory', color='gray', \
         horizontalalignment='left', transform=ax30.transAxes)


plt.savefig(std_name + '_obs-conditions-plots.png', dpi=72)

plt.close("all")  # clears memory



##################################################################################################
# Write the magdev.txt file with mean delta mag (against VAR) and std dev, and
# Calculate v-c for all stars and store in a dataframe
##################################################################################################

print('Writing magdev.txt file.\n')
min_good = max(int(0.1 * frm), 2)      # the minimum number of good points to make it to the file

f_magdev = open('magdev.txt', 'w')
f_magdev.write('INDEX MEAN_MAG STDEV GOODPOINTS\n')

comp_df = df2_list[idxc]    # the dataset for COMP
var_df = df2_list[idxv]    # the dataset for VAR
varmeanmag = (var_df['mag'] - comp_df['mag']).mean()

all_vc_list = []  # will hold a list of V-C stats for all stars, [id, mean, err, good-pts] each

for idx in range(len(all_stars)):
    star_id = all_stars[idx]
    star_vc_mag = df2_list[idx]['mag'] - comp_df['mag']
    star_mean = star_vc_mag.mean() - varmeanmag         # zero point of the magdev chart is VAR
    star_sd = star_vc_mag.std()
    star_good = star_vc_mag.count()
    all_vc_list.append([star_id, star_mean, star_sd, star_good])

    if star_good > min_good:
       f_magdev.write(star_id + ' ' + str(round(star_mean,5)) + ' ' +
                      str(round(star_sd,5)) + ' ' + str(star_good) + '\n')

vc_stat_df = pd.DataFrame(all_vc_list, columns=['id', 'mean', 'sd', 'good-pts'])
vc_stat_df.set_index('id', inplace=True)
vc_stat_df_sort = vc_stat_df.sort_values('mean')
f_magdev.close()


####################################################################################################
#
# Add to the selected stars distributing them in all 4 quadrants of the image around the centre
#
####################################################################################################

# find the max numbering of the chk stars that have already been selected
max_chk_no = 0
for s_label in sel_df.index.values:
    if 'chk' in s_label:
        chk_no_str = re.search(r'chk([0-9]*)', s_label).group(1)
        if chk_no_str == '':  # covers the case where the label is just 'chk'
            max_chk_no = 1
        else:
            chk_no = int(chk_no_str)
            if chk_no > max_chk_no:
                max_chk_no = chk_no


# do not add stars if there are already more than the min no of stars
if max_chk_no + 2 < min_stars:

    # calc the max radius from the centre of the CCD to select stars, ie keep away from edges
    centre_x = float(framex)/2
    centre_y = float(framey)/2

    ccd_width = float(framey)
    if float(framex) < float(framey):  # in case x dimension is smaller than y
        ccd_width = float(framex)
    max_rad_sel = 1.0 * ccd_width/2

    good_threshold = 0.8 * frm  # must be at least this number of good points to be selected

    # a list of 4 empty lists, each of which will hold the ids of selected stars in each quarter
    sel_by_q = [[] for i in range(4)]
    q_count = [0 for i in range(4)]   # count of the number of stars that have been selected in each quarter


    # select stars in order of brightness
    for s_id in vc_stat_df_sort.index.values:  # for each star id
        # must have the threshold no of good points and not been selected before
        if int(vc_stat_df.loc[s_id]['good-pts']) > good_threshold and s_id not in sel_df['id'].values:
            s_id_x = float(star_df.loc[s_id]['x'])
            s_id_y = float(star_df.loc[s_id]['y'])

            rad_out = math.sqrt((s_id_x - centre_x)**2 + (s_id_y - centre_y)**2)
            if rad_out < max_rad_sel:
                if s_id_x > centre_x:
                    if s_id_y > centre_y:
                        if q_count[0] < add_stars/4:  # do not add any more stars to this quarter if limit exceeded
                            sel_by_q[0].append(s_id)   # this star is in quarter 0
                            q_count[0] += 1
                    else:
                        if q_count[1] < add_stars/4:
                            sel_by_q[1].append(s_id)   # this star is in quarter 1
                            q_count[1] += 1
                else:
                    if s_id_y > centre_y:
                        if q_count[2] < add_stars/4:  # do not add any more stars to this quarter if limit exceeded
                            sel_by_q[2].append(s_id)   # this star is in quarter 2
                            q_count[2] += 1
                    else:
                        if q_count[3] < add_stars/4:
                            sel_by_q[3].append(s_id)   # this star is in quarter 3
                            q_count[3] += 1
        tot_add = q_count[0]+q_count[1]+q_count[2]+q_count[3]
        if tot_add > add_stars or max_chk_no + 2 + tot_add >= star_limit:
            break  # stop once we have selected enough stars

    added_ids = []
    for i in range(4):
        for s in sel_by_q[i]:
            added_ids.append(s)  # converts all added selected stars into a single list


    added_stars = []
    for i in range(tot_add):
        if (max_chk_no + i + 1) == 1:
            new_label = 'chk'
        else:
            new_label = 'chk' + str(max_chk_no + i + 1)
        added_stars.append([new_label, added_ids[i]])

    added_df = pd.DataFrame(added_stars, columns=['label', 'id'])
    added_df.set_index('label', inplace=True)

    sel_df = sel_df.append(added_df)  # the additional stars are appended to the original selection

    print('\nAdded selection stars: ' + str(tot_add))
    ####################################################
    # these additional stars are added to the ref xml
    ####################################################

    xml_lines = open(catalog_dir + target_name + '.xml').readlines()
    f_xml = open(catalog_dir + target_name + '.xml', 'w')  # open to re-write the xml
    for line in xml_lines:
        if line.strip() != '</selection>':
            f_xml.write(line)  # not yet at the end of the old selection list so just re-output the lines unchanged
        else:
            for i in range(tot_add):
                new_line = '<select id="' + str(added_stars[i][1]) + '" label="' + added_stars[i][0] + '"/>\n'
                f_xml.write(new_line)
            f_xml.write(line)  # this should be just '</selection>' after which the rest of the file is written as is




    f_xml.close()


else:
    print('\nDid not add any stars because there are already '+ str(max_chk_no+2) +
          ' selected stars \n- above the lower limit of ' + str(min_stars) + '.\n')

f_select = open('select_labels.txt', 'w')  # file with a list of selection labels and star ids
f_select.write(sel_df.to_string())
f_select.close()



###################################################################################################
# EBCheck code inserted from here
###################################################################################################



if check_gaia:

    #jd_obs = init_jd
    print('\n\n*** Observation JD to be used for proper motion calcs = ' + str(init_jd))

########################################################################################################


if os.path.exists(catalog_dir + target_name + ".xml"):
    f_in_lines = open(catalog_dir + target_name + ".xml").readlines()  # read all lines in catalog file

    if os.path.exists(magdev_filename):
        magdev_lines = open(magdev_filename).readlines()    # read all lines


    else:
        print("Error:  magdev.txt not found")
        sys.exit()

    out_f = open(target_name + out_filename_template, 'w')      # overall output data
    out_f.write('#id\tdist\tdel_mag\tstdev\tgoodpts\tmax_sig\tx-coord\ty-coord\n')

    ##########################################################################################################
    # go through the file once to identify the VAR and COMP stars and pick up situations with multiples of these
    # Build list of star labels, which will be matched later with x, y coords

    label_list = []   # empty lists to hold star label, id
    id_list = []

    for line in f_in_lines:
        expr = r"<select\s+id=\"([0-9]+)\"\s+label=\"([\w]+)\""
        if re.search(expr, line):   # this is the section that contains labels for selected stars
            sub_line = re.search(expr, line)
            star_id = sub_line.group(1)
            star_label = sub_line.group(2)
            #if star_label == 'chk':         # this goup of statements aligns star numbering with the LC plots
            #    star_label = 'chk#1'        # Muniwin uses chk instead of chk#1 - makes label clearer
            if star_label == 'var':         # mark that VAR and COMP have special meanings - not just ensemble stars
                star_label = 'VAR ('+star_id+')'
            elif star_label =='comp':
                star_label = 'COMP ('+star_id+')'
            else:
                star_label = star_id
                #sub_star_id = re.search(r"chk([0-9]+)", star_label)
                #star_label = 'chk'+'#'+sub_star_id.group(1)


            label_list.append(star_label)
            id_list.append(star_id)

        if 'label="var"' in line:       # this line contains the id of the target star
            var_id = star_id
            var_count += 1

        elif 'label="comp"' in line:       # this line contains the id of the comp star
            comp_id = star_id
            comp_count += 1
            # find the mag of var against comp; magdev mags are vs VAR
            for line2 in magdev_lines:
                line_elements = line2.split()       # splits the line using any whitespace as separator

                if line_elements[0] == comp_id:
                    var_vs_comp = 0.0 - float(line_elements[1])  # VAR mag minus COMP mag

    label_array = np.array(label_list)      # convert to array for use by APLPy
    no_stars = len(label_list)      # the number of selected stars
    x_array = np.zeros(no_stars)
    y_array = np.zeros(no_stars)    # create zeroed arrays for the x and y coords of the selected stars

    if var_count != 1:
        print("ERROR: Either no VAR found, or multiple")
        sys.exit()
    if comp_count != 1:
        print("ERROR: Either no COMP found, or multiple")
        sys.exit()

    ###################################################################################################
    #   go through file a 2nd time to assign var and comp mags, var coordinates

    for line in f_in_lines:
        # the \s+ in the regular expression matches an arbitrary no of white spaces between items
        expr = r"<s\s+id=\"([0-9]+)\"\s+x=\"([.0-9]+)\"\s+y=\"([.0-9]+)\""

        if re.search(expr, line):        # section of the catalog file that lists all stars detected
            sub_line = re.search(expr, line)
            star_id = sub_line.group(1)
            star_x = sub_line.group(2)
            star_y = sub_line.group(3)
            #star_mag = sub_line.group(4)        # this mag is not reliable - use only for plotting
            if star_id == var_id:
                var_x = star_x
                var_y = star_y
                vm = var_vs_comp + ref_cm
                #var_mag = str(vm)

                vx = float(var_x)
                vy = float(var_y)

                v_flux = 10.0**(10.0 - vm/2.5)    # reference intensity used by Muniwin is 10^10, not reliable

            #elif star_id == comp_id:
            #    comp_mag = star_mag
            #    cm = float(comp_mag)


    ##############################################################################################
    # go through file a 3rd time to identify stars that are closer than proximate_dist away
    # also populate the arrays containing x and y coords of selected stars

    # create empty lists to hold star-id, mag, sd, to be used for plotting
    sel_id = []
    sel_mvv = []
    sel_sd = []

    for line in f_in_lines:
        if re.search(expr, line):        # section of the catalog file that lists the stars
            sub_line = re.search(expr, line)
            star_id = sub_line.group(1)
            star_x = sub_line.group(2)
            star_y = sub_line.group(3)
            #star_mag = sub_line.group(4)

            sx = float(star_x)
            sy = float(star_y)
            #sm = float(star_mag)
            dist = scale * math.sqrt((sx - vx)**2.0 + (sy - vy)**2.0)

            if star_id in id_list:              # this star is amongst the selected!
                idx = id_list.index(star_id)
                x_array[idx] = sx               # populate their coords
                y_array[idx] = sy

            if dist <= proximate_dist*60.:      # this star is within proximate_dist of var

                # if this is an outlier copy the lightcurve to the working directory and give it a std name
                #if star_id in outlier_list:
                #    copyfile(os.path.join(in_path,'photcorr_plots','outlier-star_'+star_id+'.png'),
                #             os.path.join(in_path, std_name+'_outlier-star_'+star_id+'.png'))


                # Look up the star's delta mag and scatter in magdev.txt
                # There may also be stars in the catalog file but not in magdev.txt.
                # In those cases the stars are likely spurious so do not put on the star chart, nor output file

                star_found = False      # flags when star_id is either comp, or found in the magdev file

                for line2 in magdev_lines:
                    line_elements = line2.split()       # splits the line using any whitespace as separator

                    if line_elements[0] == star_id:
                        star_found = True
                        mag_vs_var = line_elements[1]
                        stdev = line_elements[2]
                        goodpoints = line_elements[3]

                if star_found:

                    mvv = float(mag_vs_var)
                    sd = float(stdev)

                    sel_id.append(star_id)
                    sel_mvv.append(mvv)
                    sel_sd.append(sd)

                    ###############################################################################
                    # calculate the max signal that a full eclipse in this star could cause in the target
                    # assumes all of the flux from this star is captured in the photometry aperture
                    # ie no weighting and ignores the KELT 3 arcmin aperture
                    #
                    # At the moment this calc is pretty dodgy is is of limited usefulness
                    # an area for improvement in the future
                    # ************ todo remove need for the out_file
                    ################################################################################

                    max_signal = 0      # as of V2.5 a dummy variable just set to zero


                    out_f.write(star_id+'\t'+str(round(dist/60,3))+'\t'+str(mvv)+'\t'+str(round(sd,3))\
                                +'\t'+goodpoints+'\t'+str(round(max_signal,3))+\
                                '\t'+str(sx)+'\t'+str(sy)+'\n')




else:
    print("\nError: No catalog file found.")
    sys.exit()

out_f.close()

#################################################################################
#
# this section to plate-solve and make field images
#
#################################################################################
if plate_solve:

    # plate solve the full field and make an image

    # clear the file that marks that a solve has been successful so we can test later is solve has worked
    if os.path.exists(target_name + '.solved'):
        os.remove(target_name + '.solved')

    ref_fits_filename = catalog_dir + target_name + '.fts'
    copyfile(ref_fits_filename, os.path.join(in_path,target_name + '.fts'))  # copy the ref fits file to here


    ####################################################################################
    # this call is still dependent on linux
    ####################################################################################

    # we solve the full field twice. This first one for the ref field is necessary to match x,y coords
    # read from the catalog file.  The 2nd one, of the best-frame produced the session finder image

    subprocess.call(['solve-field', '-N', target_name+'_'+observatory+'_WCS.fts', '--overwrite', '--resort', \
                     '--scale-units', 'arcsecperpix', '--scale-low', scale_low, '--scale-high', scale_high, \
                     '--downsample', '2', '--no-plots', target_name + '.fts']) # remove '--no-plots' to debug

    #####################################################################################

    if os.path.exists(target_name + '.solved'):     # solve was successful

        # plate solve produces the file target_name_<observatory>_WCS.fts
        hdulist = fits.open(target_name+'_'+observatory+'_WCS.fts')
        hdr = hdulist[0].header

        # get WCS info from the FITS file
        w = wcs.WCS(hdr)


        field = aplpy.FITSFigure(target_name+'_'+observatory+'_WCS.fts', north=rotate, figsize=(15,10))

        # in some cases pmin may have to be increased from default of 0.25
        # or if pixel values are significantly -ve, force vmin to be slightly less than sky level
        try:
            field.show_grayscale(pmin=0.25, stretch='log', invert='True')
        except:
            field.show_grayscale(vmin=best_frame_sky*0.95, stretch='log', invert='True')
            #print("\n***WARNING***  Default field plot setting in line:")
            #print("  field.show_grayscale(pmin=0.25, stretch='log', invert='True'")
            #print("  did not work. Setting vmin to 110. Reduce if plot background is whited out.\n")

        field.add_grid()
        field.grid.set_color('blue')
        field.grid.set_alpha(0.2)

        field.set_title('Reference Field: ' + target_name)
        field.add_scalebar(0.08333)    # 5 arcmin in degrees
        field.scalebar.set_corner('bottom left')
        field.scalebar.set_font(size='medium', weight='bold')
        field.scalebar.set_linewidth(2)
        field.scalebar.set_label('5 arcmin')
        field.add_label(0.0, -0.07, 'Image: '+observatory+' Observatory', relative=True, color='grey',
                        weight='light', size='small')


        #  calculate positions and place labels

        world = w.all_pix2world(x_array, y_array, 0)      # 0 is the origin of the coord syst (not 1)
        x_world = world[0]
        y_world = world[1]
        field.show_markers(x_world, y_world, color='maroon')

        for idx in range(0, no_stars):
            field.add_label(x_world[idx]-0.01, y_world[idx]+0.006, label_array[idx], \
                            stretch='extra-expanded', color='blue', weight='roman')

        field.save(target_name + '_ref-field.png')       # save an image of the full field

        # clean up by removing unnecessary files
        if os.path.exists(target_name + '.axy'):
            os.remove(target_name + '.axy')
        if os.path.exists(target_name + '.corr'):
            os.remove(target_name + '.corr')
        if os.path.exists(target_name + '.match'):
            os.remove(target_name + '.match')
        if os.path.exists(target_name + '.rdls'):
            os.remove(target_name + '.rdls')
        if os.path.exists(target_name + '.solved'):
            os.remove(target_name + '.solved')
        if os.path.exists(target_name + '.wcs'):
            os.remove(target_name + '.wcs')
        if os.path.exists(target_name + '-indx.xyls'):
            os.remove(target_name + '-indx.xyls')

        if os.path.exists(target_name+'_'+observatory+'_WCS.fts'):
            os.remove(target_name+'_'+observatory+'_WCS.fts')
        if os.path.exists(os.path.join(in_path,target_name + '.fts')):
            os.remove(os.path.join(in_path,target_name + '.fts'))


        #########################################################################################
        # Plate solve the best_frame.fts (typical image from this observing run) image
        #########################################################################################

        # clear the file that marks that a solve has been successful so we can test later if solve has worked
        if os.path.exists('best_frame.solved'):
            os.remove('best_frame.solved')

        subprocess.call(['solve-field', '-N', std_name + '_WCS.fts', '--overwrite', '--resort', \
                     '--scale-units', 'arcsecperpix', '--scale-low', scale_low, '--scale-high', scale_high, \
                     '--downsample', '2', '--no-plots', 'best_frame.fts']) # remove '--no-plots' to debug

        if os.path.exists('best_frame.solved'):     # solve was successful

            # plate solve produces the file std_name_WCS.fts
            hdulist2 = fits.open(std_name + '_WCS.fts')
            hdr2 = hdulist2[0].header

            # get WCS info from the FITS file
            w2 = wcs.WCS(hdr2)

            field2 = aplpy.FITSFigure(std_name + '_WCS.fts', north=rotate, figsize=(15,10))
            # in some cases pmin may have to be increased from default of 0.25
            try:
                field2.show_grayscale(pmin=0.25, stretch='log', invert='True')
            except:
                field2.show_grayscale(vmin=best_frame_sky*0.95, stretch='log', invert='True')
                #print("\n***WARNING***  Default field plot setting in line:")
                #print("  field2.show_grayscale(pmin=0.25, stretch='log', invert='True'")
                #print("  did not work. Setting vmin to 110. Reduce if plot background is whited out.\n")

            field2.add_grid()
            field2.grid.set_color('blue')
            field2.grid.set_alpha(0.2)

            field2.set_title('Field: ' + target_name + ' on ' + obs_date + ' from ' + observatory + ' in ' + filter)
            field2.add_scalebar(0.08333)    # 5 arcmin in degrees
            field2.scalebar.set_corner('bottom left')
            field2.scalebar.set_font(size='medium', weight='bold')
            field2.scalebar.set_linewidth(2)
            field2.scalebar.set_label('5 arcmin')
            field2.add_label(0.0, -0.07, 'Image: '+observatory+' Observatory', relative=True, color='grey',
                             weight='light', size='small')


            #  place markers and labels

            field2.show_markers(x_world, y_world, color='maroon')

            for idx in range(0, no_stars):
                field2.add_label(x_world[idx]-0.007, y_world[idx]+0.006, label_array[idx], \
                            stretch='extra-expanded', color='blue', weight='roman')

            field2.save(std_name + '_field.png')       # save an image of the full field

            # clean up by removing unnecessary files
            if os.path.exists('best_frame' + '.axy'):
                os.remove('best_frame' + '.axy')
            if os.path.exists('best_frame' + '.corr'):
                os.remove('best_frame' + '.corr')
            if os.path.exists('best_frame' + '.match'):
                os.remove('best_frame' + '.match')
            if os.path.exists('best_frame' + '.rdls'):
                os.remove('best_frame' + '.rdls')
            if os.path.exists('best_frame' + '.solved'):
                os.remove('best_frame' + '.solved')
            if os.path.exists('best_frame' + '.wcs'):
                os.remove('best_frame' + '.wcs')
            if os.path.exists('best_frame' + '-indx.xyls'):
                os.remove('best_frame' + '-indx.xyls')
            if os.path.exists('best_frame.fts'):
                os.remove('best_frame.fts')


        else:
            print('\nERROR: best_frame.fts did not solve\n')


        #########################################################################################
        #  Make the neighbourhood info file, using WCS from the FITS and x,y from ...ebcheck.txt
        #########################################################################################

        # create lists that will hold RA, DEC coords and star id
        zid_list = []
        zx_list = []
        zy_list = []
        zmag_list = []

        # read the ...ebcheck.txt file to get star info, and open neighbourhood file for writing
        ebcheck_lines = open(target_name + out_filename_template).readlines()  # read all lines
        first_line = ebcheck_lines.pop(0)   # discard first line which is the header

        ref_vm = ref_cm + var_vs_comp       # correctly referenced mag of VAR
        var_flux = 10.0**(10.0 - ref_vm/2.5)

        hood_f = open(hood_filename, 'w')
        hood_f.write('# Neighbourhood and NEB Clearance Report\n\n')
        hood_f.write('# Stars within '+str(round(proximate_dist,1))+' arcmin of '+target_name+', star-id #'+var_id+'\n')
        hood_f.write('# Magnitudes are referenced to COMP = '+str(ref_cm)+'\n')
        hood_f.write('# Fluxes are referenced to VAR = 1000\n')
        hood_f.write('# dmag is referenced to VAR, dist is from target.\n')
        hood_f.write('# Exposure time (s) = '+exptime+'\n\n')

        hood_f.write("# For stars within " + str(proximate_dist) + "' of VAR, report the following as well:\n")
        hood_f.write('# norm SD for 120s, NEB depth calculated using dmag-'+ str(dmag_buf) +
                     'mag, allowing for any diff to TESS dmag.\n')
        hood_f.write('# Expected transit depth = '+str(round(TESSdepth,3))+'ppt or '
                     +str(round(TESSdepth_mmag,3))+'mmag\n\n')

        hood_f.write('#id\tRA (J2000) \tDEC (J2000) \tflux \tmag\tSD\tvalid\tdmag\tdist\tnorm\t'
                     'NEB\tB/A\tDisposition\n') #header line
        hood_f.write('\t\t\t\t\t\t\t(mag)\tpoints\t\t\tSD(A)\tdepth(B)\t\t\n') #header line
        hood_f.write('\t\t\t\t\t\t\t\t\t\t\t(ppt)\t(ppt)\t\t\n\n')

        # prepare to add selected proximate star LCs to a zip file
        prox_zip = zipfile.ZipFile(std_name + '_proximate.zip', 'w')
        zipped = 0  # counter for no of files zipped

        print('\nPlotting proximate star lightcurves.', end='', flush=True)

        max_mag = 0
        for line in ebcheck_lines:
            line_elements = line.split()        # separator is any whitespace
            star_id = line_elements[0]

            dist_arcmin = line_elements[1]
            dist_min = math.floor(float(dist_arcmin))
            dist_sec = round((float(dist_arcmin) - dist_min)*60.)
            dist_minsec = str(dist_min) + "'" + str(dist_sec) + '"'  # express in min'sec"

            dmag = float(line_elements[2])
            mag = dmag + ref_vm      # convert to standard mag scale
            sd = line_elements[3]
            norm_sd = 1000*float(sd)/math.sqrt(120.0/float(exptime))
            norm_sd_ppt = 1000*(1 - 10**(-norm_sd/1000/2.5))  # convert from mmag to ppt
            goodpoints = line_elements[4]

            img_x = float(line_elements[6])
            img_y = float(line_elements[7])

            if mag > max_mag:
                max_mag = mag

            pixcrd = np.array([[img_x, img_y]], np.float)
            world = w.all_pix2world(pixcrd, 0)      # 0 is the origin of the coord syst (not 1)

            ra = Angle(world[0,0], u.degree)
            dec = Angle(world[0,1], u.degree)

            zid_list.append(star_id)
            zx_list.append(world[0,0])
            zy_list.append(world[0,1])
            zmag_list.append(mag)

            # note the world coords of VAR
            if star_id == var_id:
                var_world = world
                var_ra = ra
                var_dec = dec


            flux = 10.0**(10.0 - mag/2.5)

            out_str = star_id+'\t'+ra.to_string(unit=u.hour, precision=2)+\
                      '\t'+ dec.to_string(unit=u.degree, precision=1)+'\t'+str(round(flux*1000.0/var_flux))+\
                      '\t'+str(round(mag,2))+'\t'+sd+'\t'+ goodpoints +'\t'+str(round(dmag,2)) +'\t'+dist_minsec

            # if the star is within proximate_dist (set by discussion with the TESS team) of the target print the additional info
            if float(dist_arcmin) <= proximate_dist:
                NEBdepth = TESSdepth/(10**(-(dmag-dmag_buf)/2.5))  #dmag_buf allows for diff between Tmag and Obs-mag
                NEBdepth_str = str(round(NEBdepth))

                if NEBdepth > 1000:  # flux cannot be negative, so this tells plot_lc not to plot the bottom
                    NEBmag = -1
                    #NEBmag_str = 'NA'
                else:
                    NEBmag = -2.5*math.log10(1.0-NEBdepth/1000)
                    #NEBmag_str = str(round(NEBmag*1000))

                if star_id == comp_id:
                    confidence = 0
                    confidence_str = 'NA'
                    disposition = 'COMP star'
                else:
                    confidence = NEBdepth/norm_sd_ppt
                    confidence_str = str(round(confidence,1))
                    if star_id == var_id:
                        disposition = 'Target star'
                    else:
                        if NEBdepth > 1000:
                            #confidence_str = 'NA'
                            disposition = 'Cleared - too faint'
                        elif confidence > 5.0:
                            disposition = 'Cleared'
                        elif confidence > 3.0:
                            disposition = 'Likely cleared'
                        else:
                            disposition = '*** Not Cleared ***'

                        ####################################################################################
                        # Plot LC and NEBDepth
                        ####################################################################################
                        plot_filename = std_name + '_proximate_' + star_id + '.png'

                        idx = np.where(all_stars == str(star_id))[0][0]
                        star_df = df2_list[idx]  # dataset for currently selected star

                        vc_mag = star_df[['mag']] - comp_df[['mag']]     # V-C
                        vc_err_sq = star_df[['err']]**2 + comp_df[['err']]**2
                        vc_err = np.sqrt(vc_err_sq['err'])
                        vc_df = pd.concat([star_df[['bjd']],vc_mag, vc_err], axis=1) # make new df with results

                        bjd = vc_df['bjd'].values
                        mag = vc_df['mag'].values + ref_cm
                        err = vc_df['err'].values

                        meanmag = np.nanmean(mag)
                        std = np.nanstd(mag, ddof=1)

                        title = 'Proximate star Star ID: ' + star_id
                        print('.', end='', flush=True)
                        plot_lc(bjd,mag, err, meanmag, std, title, plot_filename,
                                ingress, egress, NEBmag, NEBdepth, disposition)

                        if disposition == 'Cleared' or disposition == 'Likely cleared' or \
                                disposition == '*** Not Cleared ***':
                            prox_zip.write(plot_filename, compress_type=zipfile.ZIP_DEFLATED)
                            zipped += 1

                        ####################################################################################


                out_str += '\t'+ str(round(norm_sd_ppt)) + \
                           '\t'+ NEBdepth_str + '\t'+ confidence_str + '\t'+ disposition

            hood_f.write(out_str+'\n')

        zid_array = np.array(zid_list)
        zx_array = np.array(zx_list)
        zy_array = np.array(zy_list)        # create the arrays that will be used in the zoomed in plot

        prox_zip.close()
        if zipped == 0:  # remove an empty zip file
            os.remove(std_name + '_proximate.zip')
        print('\n')


        ###############################################################################################
        # Query for Gaia stars close to the target and plot on zoomed chart
        #
        ###############################################################################################

        if check_gaia:  # todo - put in a try except to catch when there isn't an internet connection

            # create lists that will hold Gaia star id, RA, DEC coords
            gaia_id_list = []
            gaia_ra_list = []
            gaia_dec_list = []
            gaia_rmag_list = []
            gaia_gmag_list = []

            # these lists will hold results of matching between Gaia stars and detected stars
            theta_to_var_list = []
            id_found_list = []
            min_theta_found_list = []
            found_delta_mag_list = []
            found_star_id_list =[]



            # calculate mag of the brightest star that could have caused the TESS signal
            TESS_dmag = -2.51188643150958*np.log10(TESSdepth/1000.)
            print('\nChecking for stars within dmag = '+str(round(TESS_dmag,1))+'\n\n')


            # set up proper motion corrections from Gaia values in milli-arcsecs per year
            pm_scale = 1./1000.

            # query Gaia database
            var_coord = coordinates.SkyCoord(var_ra, var_dec, unit=(u.degree, u.degree), frame='icrs')
            radius = u.Quantity(proximate_dist/60., u.degree)

            gaia_table = Gaia.query_object_async(coordinate=var_coord, radius=radius)

            n_gaia_stars = len(gaia_table)
            if n_gaia_stars < 1:
                print('\n********* No Gaia objects found within the specified radius. ********\n\n')
            else:

                hood_f.write('\n#**************************************************************************************\n')
                hood_f.write('# Gaia star check section\n\n')
                hood_f.write('# Checking for stars within dmag = '+str(round(TESS_dmag,1))+" plus 'slack'.\n")
                hood_f.write("# This slack = "+str(round(dmag_buf,1))+" for r. If r not found use g with an additional 0.5\n\n")

                hood_f.write('# err_d is the angular separation between the detected star and the Gaia star.\n')
                hood_f.write('# err_gm is the delta between the measured mag of the detected star and Gaia g mag.\n')
                hood_f.write('#**************************************************************************************\n\n')

                hood_f.write('#Gaia id\t\tRA (J2000) \tDEC (J2000) \tr_mag \tg_mag\tdist\terr_d\terr_gm\tDisposition\n\n')

                var_found = False
                var_gaia_rmag = ref_vm  # if for any reason we don't find the star corresponding to VAR these are defaults
                var_gaia_gmag = ref_vm
                var_gaia_ra = var_ra
                var_gaia_dec = var_dec

                for i in range(n_gaia_stars):
                    ra_gaia = gaia_table['ra'][i]
                    dec_gaia = gaia_table['dec'][i]
                    src_id = gaia_table['source_id'][i]
                    mag_g = gaia_table['phot_g_mean_mag'][i]
                    mag_b = gaia_table['phot_bp_mean_mag'][i]
                    mag_r = gaia_table['phot_rp_mean_mag'][i]
                    ref_epoch = gaia_table['ref_epoch'][1]
                    pm_ra =  gaia_table['pmra'][i]
                    pm_dec = gaia_table['pmdec'][i]
                    distance = gaia_table['dist'][i]

                    # Invalid entries are numpy.ma.core.MaskedConstant
                    # Valid entries are floating point
                    # Test for invalid entries and skip them
                    # In diagnostics a print(variable) will show "--" for a MaskedConstant

                    # Update coordinates for proper motion to epoch of observation
                    # Proper motion is in mas/yr
                    # init_jd is in Julian days
                    # ref_epoch is in calendar years
                    # ra and dec are in degrees

                    # Find the jd for the epoch of Gaia data
                    jd_ref = jd_for_epoch(ref_epoch)

                    # Find elapsed time since Gaia epoch for this observation in years
                    delta_t = (init_jd - jd_ref) / 365.25

                    # Add proper motion in degrees to the Gaia coordinates of epoch

                    if isinstance(pm_ra, np.ma.core.MaskedConstant) or isinstance(pm_dec, np.ma.core.MaskedConstant):
                      delta_ra = 0.
                      delta_dec = 0.
                    else:
                      delta_ra = pm_ra * pm_scale * delta_t
                      delta_dec = pm_dec * pm_scale * delta_t

                    ra = ra_gaia + (delta_ra/3600.)
                    dec = dec_gaia + (delta_dec/3600.)

                    ra_angle = Angle(ra, u.degree)
                    dec_angle = Angle(dec, u.degree)

                    # Entries are already selected for  separation from target based on Gaia coodinates
                    # Only include those within this distance that can affect the TESS photometry
                    # This also removes many (but not all) entries with NaN's for mag_r or mag_g

                    # Trap missing mag_r or mag_g entries by setting them unrealistically faint
                    # This will suppress them so they do not affect a sort to find the target star
                    # It still leaves them in the Gaia search table

                    if isinstance(mag_r, np.ma.core.MaskedConstant):
                      mag_r = 99.
                    if isinstance(mag_g, np.ma.core.MaskedConstant):
                      mag_g = 99.



                    ############################################################################
                    # Find the Gaia star that corresponds to VAR, and so use its r_mag and g_mag
                    # to decide if other Gaia stars are bright enough to have caused the signal
                    ############################################################################
                    min_theta_found = max_theta
                    found_star_id = '0'
                    for i in range(len(zid_list)):
                        theta = delta_theta(zx_list[i], zy_list[i], ra, dec)

                        if min_theta_found > theta and abs(mag_g - zmag_list[i]) < max_delta_mag:
                            # found a star within the limit
                            min_theta_found = theta
                            found_star_id = zid_list[i]

                    if found_star_id == var_id:  # found VAR
                        var_found = True
                        var_gaia_rmag = mag_r
                        var_gaia_gmag = mag_g
                        var_gaia_ra = ra
                        var_gaia_dec = dec
                        break  # no need to continue to check other Gaia stars

                if var_found:
                    print('\nFound Gaia star corresponding to VAR with rmag and gmag: '+str(round(var_gaia_rmag,1))+
                          ' '+str(round(var_gaia_gmag,1))+'\n')
                else:
                    print("\nSomething's gone wrong. Did not find any Gaia star corresponding to VAR!")
                    print('This can happen if VAR is saturated and the selection is on the flux bleed.')
                    print('Or if max_theta in the pest_pipeline_config.txt file is set too small.')
                    sys.exit()


                ##################################################################################
                # Now that we have found the gaia mags of VAR, we go through again to select stars
                ##################################################################################


                for i in range(n_gaia_stars):
                    ra_gaia = gaia_table['ra'][i]
                    dec_gaia = gaia_table['dec'][i]
                    src_id = gaia_table['source_id'][i]
                    mag_g = gaia_table['phot_g_mean_mag'][i]
                    mag_b = gaia_table['phot_bp_mean_mag'][i]
                    mag_r = gaia_table['phot_rp_mean_mag'][i]
                    ref_epoch = gaia_table['ref_epoch'][1]
                    pm_ra =  gaia_table['pmra'][i]
                    pm_dec = gaia_table['pmdec'][i]
                    distance = gaia_table['dist'][i]

                    # Invalid entries are numpy.ma.core.MaskedConstant
                    # Valid entries are floating point
                    # Test for invalid entries and skip them
                    # In diagnostics a print(variable) will show "--" for a MaskedConstant

                    # Update coordinates for proper motion to epoch of observation
                    # Proper motion is in mas/yr
                    # init_jd is in Julian days
                    # ref_epoch is in calendar years
                    # ra and dec are in degrees

                    # Find the jd for the epoch of Gaia data
                    jd_ref = jd_for_epoch(ref_epoch)

                    # Find elapsed time since Gaia epoch for this observation in years
                    delta_t = (init_jd - jd_ref) / 365.25

                    # Add proper motion in degrees to the Gaia coordinates of epoch

                    if isinstance(pm_ra, np.ma.core.MaskedConstant) or isinstance(pm_dec, np.ma.core.MaskedConstant):
                      delta_ra = 0.
                      delta_dec = 0.
                    else:
                      delta_ra = pm_ra * pm_scale * delta_t
                      delta_dec = pm_dec * pm_scale * delta_t

                    ra = ra_gaia + (delta_ra/3600.)
                    dec = dec_gaia + (delta_dec/3600.)

                    # calc the separation from VAR
                    theta_to_var = delta_theta(var_gaia_ra, var_gaia_dec, ra, dec)


                    # Entries are already selected for  separation from target based on Gaia coodinates
                    # Only include those within this distance that can affect the TESS photometry
                    # This also removes many (but not all) entries with NaN's for mag_r or mag_g

                    # Trap missing mag_r or mag_g entries by setting them unrealistically faint
                    # This will suppress them so they do not affect a sort to find the target star
                    # It still leaves them in the Gaia search table

                    if isinstance(mag_r, np.ma.core.MaskedConstant):
                      mag_r = 99.
                    if isinstance(mag_g, np.ma.core.MaskedConstant):
                      mag_g = 99.


                    # now use the gaia mags to select stars that could have caused the signal


                    if (mag_r > 0. and mag_r < 22.):
                      selected_flag = (mag_r < (var_gaia_rmag + TESS_dmag + dmag_buf))
                    elif (mag_g > 0. and mag_g < 22.):
                      selected_flag = (mag_g < (var_gaia_gmag + TESS_dmag + dmag_buf + 0.5)) # not sure if the extra 0.5 is needed
                    else:
                      selected_flag = False

                    if selected_flag:   # build lists of selected stars for plotting
                        gaia_id_list.append(src_id)
                        gaia_ra_list.append(ra)
                        gaia_dec_list.append(dec)
                        gaia_rmag_list.append(mag_r)
                        gaia_gmag_list.append(mag_g)

                        ###################################################################################
                        # for each Gaia star above the dmag threshold, find the #id that corresponds
                        ###################################################################################

                        id_found = False
                        min_theta_found = max_theta
                        for i in range(len(zid_list)):
                            theta = delta_theta(zx_list[i], zy_list[i], ra, dec)

                            if min_theta_found > theta and abs(mag_g - zmag_list[i]) < max_delta_mag:
                                # found a star within the limit
                                min_theta_found = theta
                                id_found = True
                                found_star_id = zid_list[i]
                                found_delta_mag = mag_g - zmag_list[i]


                        if not id_found:
                            found_star_id = 'NA'
                            found_delta_mag = 0.0

                        ##########################################################################
                        # build an initial list of selected Gaia stars
                        # will be checked for duplication in the next section
                        ##########################################################################

                        theta_to_var_list.append(theta_to_var)
                        id_found_list.append(id_found)
                        min_theta_found_list.append(min_theta_found)
                        found_delta_mag_list.append(found_delta_mag)
                        found_star_id_list.append(found_star_id)




                #########################################################################################
                #
                # find all Gaia stars assigned to the same star ids, and keep only the closest one
                #
                #########################################################################################


                for g_star in set(found_star_id_list): # each of the star ids found, ignoring duplicates
                    if g_star != 'NA':
                        dup_list = [i for i, id in enumerate(found_star_id_list) if id == g_star]
                        if len(dup_list) > 1:  # ie there are duplicates

                            dup_thetas = []  # will hold the thetas for each of the duplicate stars
                            for j in dup_list:
                                dup_thetas.append(min_theta_found_list[j])

                            idx_min_theta = np.argmin(dup_thetas) # returns the index of the min theta entry
                            for i in range(len(dup_list)):
                                if i != idx_min_theta:
                                    found_star_id_list[dup_list[i]] = 'NA'  # mark the gaia star as not matched
                                    id_found_list[dup_list[i]] = False


                ##########################################################################################
                #
                # this section examines the finalised star match lists and writes all that out to file
                #
                ##########################################################################################

                for i in range(len(gaia_id_list)):

                    ra_angle = Angle(gaia_ra_list[i], u.degree)
                    dec_angle = Angle(gaia_dec_list[i], u.degree)
                    mag_r = gaia_rmag_list[i]
                    mag_g = gaia_gmag_list[i]

                    dist_min = math.floor(theta_to_var_list[i]/60.)
                    dist_sec = int(round(theta_to_var_list[i] - dist_min*60.))
                    dist_minsec = str(dist_min) + "'" + str(dist_sec) + '"'  # express in min'sec"

                    if id_found_list[i]:  # marks if this particular Gaia star was matched
                        min_theta_found_txt = str(round(min_theta_found_list[i],2))+'"'
                        found_delta_mag_txt = str(round(found_delta_mag_list[i],1))
                        disposition = '== star #' + found_star_id_list[i]

                    else:
                        min_theta_found_txt = ' '
                        found_delta_mag_txt = ' '
                        disposition = '***Not Cleared - not detected***'
                        #found_star_id = 'NA'  # marks that star not matched to any detected star
                        #found_delta_mag = 0
                        #min_theta_found = 0



                    out_line = str(gaia_id_list[i])+'\t'+ra_angle.to_string(unit=u.hour, precision=2)+'\t'+\
                               dec_angle.to_string(unit=u.degree, precision=1)+'\t'+str(round(mag_r,1))+\
                               '\t'+str(round(mag_g,1))+'\t'+dist_minsec+'\t'+min_theta_found_txt+'\t'+\
                               found_delta_mag_txt+'\t'+disposition

                    hood_f.write(out_line+'\n')



                #convert to arrays as expected by APLpy
                gaia_id_array = np.array(gaia_id_list)
                gaia_ra_array = np.array(gaia_ra_list)
                gaia_dec_array = np.array(gaia_dec_list)



        hood_f.write('\n#**************************************************************************************\n\n')
        max_dmag = max_mag-ref_vm
        hood_f.write('# Max contrast: ' + str(round(max_dmag, 1)) + ' mag within ' + str(proximate_dist*60) + '"')
        hood_f.close()





        ################################################################################################
        #  make a new FITS file of just the area around the target
        #  first we find the x,y coords of VAR in best_frame.fts
        ################################################################################################

        var_xy = w2.all_world2pix(var_world, 0)

        #  first we check if the re-section would spill over the original image borders

        rad_pix = proximate_dist*60./scale      # radius from target to check, in pixels
        x1 = int(var_xy[0,0] - rad_pix)
        x2 = int(var_xy[0,0] + rad_pix)
        y1 = int(var_xy[0,1] - rad_pix)
        y2 = int(var_xy[0,1] + rad_pix)

        if x1 < 1:
            x1 = 1
        if y1 < 1:
            y1 = 1

        if x2 > int(hdr['naxis1']):
            x2 = int(hdr['naxis1'])
        if y2 > int(hdr['naxis2']):
            y2 = int(hdr['naxis2'])


        ###########################################################################################
        # another linux dependent call
        ###########################################################################################

        # this line is the command: fitscopy std_name_WCS.fts[x1:x2,y1:y2] !zoomed.fts
        # the ! is needed to force overwriting of the output file
        subprocess.call(['fitscopy',std_name+"_WCS.fts["+str(x1)+":"+str(x2)+","+str(y1)+":"+str(y2)+"]",\
                         '!zoomed.fts'])

        ###########################################################################################

        # the subimage sectioning produces the file zoomed.fts

        zoomed = aplpy.FITSFigure('zoomed.fts', north=rotate, figsize=(8,8))
        #zoomed.show_grayscale(stretch='log', invert='True')

        try:
            zoomed.show_grayscale(pmin=0.25, stretch='log', invert='True')
        except:
            zoomed.show_grayscale(vmin=best_frame_sky*0.95, stretch='log', invert='True')

        zoomed.add_grid()
        zoomed.grid.set_color('blue')
        zoomed.grid.set_alpha(0.2)

        zoomed.set_title(str(round((x2-x1)*scale/60,1))+"' x "+str(round((y2-y1)*scale/60,1))+"' field: "+target_name)
        zoomed.add_scalebar(0.016667)    # 1 arcmin in degrees
        zoomed.scalebar.set_corner('bottom left')
        zoomed.scalebar.set_font(size='medium', weight='bold')
        zoomed.scalebar.set_linewidth(2)
        zoomed.scalebar.set_label('1 arcmin')
        zoomed.add_label(0.0, -0.08, 'Image: '+observatory+' Observatory', relative=True, color='grey',
                         weight='light', size='small')
        zoomed.save(std_name + '_zoomed-BARE.png')      # for clarity this plot has no apertures nor star id

        zoomed.add_label(0.01, 0.98, 'Aperture rad = ' + str(round(phot_rad * scale, 1)) + '\"', \
                         relative=True, color='maroon', weight='medium', horizontalalignment='left')
        zoomed.add_label(0.01, 0.94, 'Large circle = '+str(round(proximate_dist,1))+"'", relative=True,
                         color='maroon', weight='medium', horizontalalignment='left')
        zoomed.show_circles(zx_array, zy_array, phot_rad_deg, edgecolor='maroon', facecolor='none')

        ####### plot Gaia star positions #########
        if check_gaia:
            zoomed.add_label(0.01, 0.90, 'Red x are Gaia stars', relative=True, color='red', weight='medium',
                             horizontalalignment='left')
            zoomed.show_markers(gaia_ra_array, gaia_dec_array, s=20, c='red', marker='x', alpha=0.8)

        # show the NEB clearance circle
        zoomed.show_circles([var_world[0,0]], [var_world[0,1]], proximate_dist/60., edgecolor='maroon',
                            facecolor='none', alpha=0.5)


        for idx in range(0, len(zid_list)): # we scale how far labels are placed from the star
            zoomed.add_label(zx_array[idx]-0.002-0.001*proximate_dist/6, zy_array[idx]+0.0, zid_array[idx], \
                            color='blue', weight='normal', horizontalalignment='left')

        zoomed.save(std_name + '_zoomed.png')      # the whole shebang


        ################################################################################
        #  End of section dependent on plate solving
        ################################################################################


# make the mag-dev plot
fig, ax1 = plt.subplots(figsize=(10,6))
plt.rc('grid', linestyle='dotted', color='gray')
ax1.set_yscale('log')

plt.plot(sel_mvv, sel_sd, 'r+')
for label, x, y in zip(sel_id, sel_mvv, sel_sd):
    plt.annotate(label, xy=(x, y), xytext=(4, 0), textcoords='offset points',
                 ha='left', va='center', color='blue')

ax1.set_xlabel('delta-mag (vs VAR)', fontsize=12)
ax1.set_ylabel('std dev', fontsize=12)
plt.title(observatory+' '+obs_date+': Stars within '+str(round(proximate_dist,1))+' arcmin of '+target_name,
          fontsize=14)
plt.savefig(std_name+'_magdev-chart.png')


###################################################################################################
###################################################################################################
# For each Selected star;
#
# 1. plot v-c
# 2. perform regression against dist_center (distance from CCD center),fwhm and airmass
# 3. write the mags with de-trending corrections to a Muni...txt file
###################################################################################################

num_sel = len(sel_df['id'].values)  # number of selected stars, including ones added by PhotCorr
#print(num_sel)


# lm_list holds LinearRegression objects for all selected stars
lm_list = [None] * sel_df.shape[0]  # an empty list with the same number of elements as selected stars

# will hold list of dataframes with regression results from each selected star
lm_df_list = []
# will hold modeled mags, and actual-modeled for each selected star
mod_mag = []
corr_mag = []

# will hold std dev of magnitudes before and after de-trending
detrend_stats = [None] * sel_df.shape[0]  # an empty list with the same number of elements as selected stars


# create the Muni.txt file and write the first 2 lines

f_muni = open('Muni_' + init_time.iso + '.txt', 'w')

if date_type == 'HJD':
    line1 = 'JDHEL V s1 C s2 '  # the start of the line. There should exist at least V and C
    line2 = 'Aperture: ' + str(sel_ap) + ', Filter: ' + filter + ', JD: heliocentric'
else:
    line1 = 'BJD V s1 C s2 '  # the start of the line. There should exist at least V and C
    line2 = 'Aperture: ' + str(sel_ap) + ', Filter: ' + filter + ', JD: barycentric'

# create a list that defines the formats of each of the column in the Muni...txt file
muni_formats = [lambda x: "%.7f" % x, lambda x: "%.5f" % x, lambda x: "%.5f" % x,
                lambda x: "%.5f" % x, lambda x: "%.5f" % x]  # for the columns created in line1

for idx in range(num_sel-2):   # since we have already written for V and C
    line1 += 'K' + str(idx+1) + ' s' + str(idx+3) + ' '
    muni_formats.extend([lambda x: "%.5f" % x, lambda x: "%.5f" % x])

line1 += 'AIRMASS'
muni_formats.append(lambda x: "%.4f" % x)

f_muni.write(line1 + '\n')
f_muni.write(line2 + '\n')


# create an empty array that will hold all the data that will be written to the Muni...txt file
# Each row is a timestep. 2 columns for each selected star plus 2 (BJD, AIRMASS)
muni_out_a = np.empty((frm, num_sel*2 + 2))
muni_out_a[:] = np.nan


heading_line = line1.replace('JDHEL', 'BJD', 1)  # necessary because even in a HJD run the script expects BJD header
muni_headings = np.array(heading_line.split())
muni_out_df = pd.DataFrame(muni_out_a, index=df2_list[0].index.values, columns=muni_headings) # using
# the first star because frame numbers should be the same across all of them

muni_out_df['BJD'] = df2_list[0]['bjd']
muni_out_df['AIRMASS'] = df2_list[0]['airmass']


# if detrend is on, make another Muni file, and dataframe to, later, populate it
if detrend:
    f_det_muni = open('det_Muni_'+init_time.iso+'.txt', 'w')

    f_det_muni.write(line1 + '\n')
    f_det_muni.write(line2 + '\n')
    det_muni_out_df = muni_out_df.copy()





sel_labels = sel_df.index.values
count = 0
for sel_star_id in sel_df['id'].values: # this is an array
    label = sel_labels[count]

    # work out the equivalent labels for the Muni...txt file

    if label == 'var':
        muni_label = 'V'
        err_label = 's1'
    elif label == 'comp':
        muni_label = 'C'
        err_label = 's2'
    elif label == 'chk':
        muni_label = 'K1'
        err_label = 's3'
    else:
        expr = r'chk([0-9]+)'
        sub_line = re.search(expr, label)
        muni_label = 'K' + sub_line.group(1)
        err_label = 's' + str(int(sub_line.group(1))+2)


    # set up a variable for Linear Regression
    lm_list[count] = linear_model.LinearRegression()

    # convert star id to index of df2_list
    idx = np.where(all_stars == str(sel_star_id))[0][0]  # remember, all_stars is ndarray

    cur_sel_df = df2_list[idx]  # dataset for currently selected star
    cur_sel_dfi = cur_sel_df.fillna(cur_sel_df.mean(), axis=0)  # condition to fill in NaN values

    vc_mag = cur_sel_df[['mag']] - comp_df[['mag']]     # V-C

    vc_err_sq = cur_sel_df[['err']]**2 + comp_df[['err']]**2
    vc_err = np.sqrt(vc_err_sq['err'])
    vc_df = pd.concat([cur_sel_df[['bjd']],vc_mag, vc_err], axis=1) # make new df with results

    bjd = vc_df['bjd'].values       # convert to numpy array
    mag = vc_df['mag'].values
    err = vc_df['err'].values
    meanmag = np.nanmean(mag)
    std = np.nanstd(mag, ddof=1)

    if detrend and detrend_plots:  # no need to make these plots if we are not detrending since nothing to compare to
        title = 'V-C Mags; Star id#' + str(sel_star_id) + ': Label ' + label
        filename = os.path.join(plots_dir, 'Diffmag_'+label+'_star_'+str(sel_star_id)+'.png')
        plot_lc_err(bjd, mag, err, meanmag, std, title, filename)

    ###########################################################################################
    #
    # Detrending section  ****todo - detrend againt total counts, and delta airmass
    #
    ###########################################################################################

    if detrend:
        print('Detrending', label)
        # calculate distance from the CCD center and add as a column to the sel_dfi dataframe

        dist_center = np.sqrt((cur_sel_dfi[['x_coord']].values-float(framex)/2)**2 +
                              (cur_sel_dfi[['y_coord']].values-float(framey)/2)**2)
        dc = dist_center.reshape(-1)    # reshape to a 1-D array

        dcf = pd.DataFrame(dc, index=cur_sel_dfi.index, columns=['dist_center'])

        # do regression against dist_center,fwhm,airmass,sky
        # use the conditioned dataframe
        params_df = pd.concat([dcf[['dist_center']], cur_sel_dfi[['fwhm']], \
                              cur_sel_dfi[['airmass']], cur_sel_dfi[['sky']]], axis=1)

        # if any of the detrend parameters are False, drop those columns
        if not detrend_fwhm:
            params_df.drop('fwhm', axis=1, inplace=True)
        if not detrend_sky:
            params_df.drop('sky', axis=1, inplace=True)
        if not detrend_airmass:
            params_df.drop('airmass', axis=1, inplace=True)
        if not detrend_rad:
            params_df.drop('dist_center', axis=1, inplace=True)



        # condition parameter data to replace NaN with mean of the column
        # or in the more extreme case where the data is all NaN, fill with zeros
        if np.count_nonzero(~np.isnan(mag)) != 0:
            vc_dfi = vc_df.fillna(vc_df.mean(), axis=0)
        else:
            vc_dfi = vc_df.fillna(0.0, axis=0)
        lm_list[count].fit(params_df, vc_dfi[['mag']])

        # put regression results into a dataframe
        lm_results = pd.DataFrame(list(zip(params_df.columns, lm_list[count].coef_[0])), \
                                  columns=['params', label])
        lm_results.set_index('params', inplace=True)
        lm_df_list.append(lm_results)

        # plot modeled LCs
        model = lm_list[count].predict(params_df)

        # mod_mag holds the corrections to be applied to do de-trending
        remodel = np.reshape(model, (-1,)) - np.nanmean(mag)  # lm.predict returns array shape (n,1)

        # do not de-trend on var - doing so would destroy any signal
        if label == 'var':
            mod_mag.append(np.zeros_like(remodel))
        else:
            mod_mag.append(remodel)


        mod_meanmag = np.nanmean(mod_mag[count])
        mod_std = np.nanstd(mod_mag[count], ddof=1)


        # this section is to plot the mag adjustments applied - comment out if not wanted

        #zero_err = np.zeros_like(mod_mag[count])   # predictions have zero error
        #title = 'Mag Adjustments; Star id#' + str(sel_star_id) + ': Label ' + label
        #filename = 'Adjustments_' + label + '_star_' + str(sel_star_id) + '.png'
        #plot_lc_err(bjd, mod_mag[count], zero_err, mod_meanmag, mod_std, title, filename)

        # plot actual minus modeled
        corr_mag.append(mag - mod_mag[count])
        corr_err = np.multiply(err, ~np.isnan(corr_mag[count])) # drop err values where cor_mag is NaN
        corr_meanmag = np.nanmean(corr_mag[count])
        corr_std = np.nanstd(corr_mag[count], ddof=1)

        if label != 'var' and detrend_plots:      # var is not detrended
            title = 'Detrended Mags; Star id#' + str(sel_star_id) + ': Label ' + label
            filename = os.path.join(plots_dir, 'Detrended_' + label + '_star_' + str(sel_star_id) + '.png')
            plot_lc_err(bjd, corr_mag[count], corr_err, corr_meanmag, corr_std, title, filename)

        # compile the before and after stats for each star
        star_stat = [label, meanmag, std, corr_std]
        detrend_stats[count] = star_stat

    ###########################################################################################
    #
    # section to write the Muni...txt file
    #
    ###########################################################################################

    # populate the output dataframe

    muni_out_df[muni_label] = df2_list[idx]['mag']
    muni_out_df[muni_label].fillna(value=99.99999, inplace=True)    # use Muniwin convention for invalid values
    muni_out_df[err_label] = df2_list[idx]['err']
    muni_out_df[err_label].fillna(value=9.99999, inplace=True)

    if detrend:
        det_muni_out_df[muni_label] = df2_list[idx]['mag'] - mod_mag[count]
        det_muni_out_df[muni_label].fillna(value=99.99999, inplace=True)
        det_muni_out_df[err_label] = df2_list[idx]['err']
        det_muni_out_df[err_label].fillna(value=9.99999, inplace=True)



    count += 1

# write the Muni..txt file

#f_muni.write(muni_out_df.to_string(header=None, index=None))

f_muni.write(muni_out_df.to_string(header=None, index=None, formatters=muni_formats))
f_muni.close()

if detrend:
    # print out the regression coefficients
    lm_df = pd.concat(lm_df_list, axis=1)
    f_lm = open('regression_coeffs.txt', 'w')
    f_lm.write(lm_df.T.to_string(header=True, index=True, float_format=lambda x: "%.6f" % x))
    f_lm.close()

    # make a txt file with before and after de-trend stats
    dstats_df = pd.DataFrame(detrend_stats, columns=['label', 'instr_mag', 'SD_before', 'SD_after'])
    dstats_df.set_index('label', inplace=True)
    f_dstats = open('detrend_stats.txt', 'w')
    f_dstats.write(dstats_df.to_string(header=True, index=True, float_format=lambda x: "%.5f" % x))
    f_dstats.close()

    # write a file with de-trend data used
    detdata_formats = [lambda x: "%.7f" % x]    # the first column is BJD, so more precision needed
    for num in range(len(params_df.columns)):
        detdata_formats.append(lambda x: "%.4f" % x)    # less precision for subsequent columns
    params_df1 = pd.concat([comp_df[['bjd']], params_df], axis=1)
    f_detdata = open(std_name + '_detrend-data.txt', 'w')
    f_detdata.write(params_df1.to_string(header=True, index=True, formatters=detdata_formats))
    f_detdata.close()

    # write the detrended version of the Muni..txt file
    f_det_muni.write(det_muni_out_df.to_string(header=None, index=None, formatters=muni_formats))
    f_det_muni.close()


##################################################################################################
# Find outliers in magdev data and plot those that are also within proximate_dist
##################################################################################################

if outlier_plots:
    print('Plotting outlier star light curves.', end='', flush=True)

    # Make a data frame to enable a linear model to be built
    magdev_frame = pd.read_csv('magdev.txt', sep="\s+", skiprows=1, header=None, engine='python')
    magdev_frame.columns = ['id', 'mag', 'sd', 'good-pts']
    magdev_frame = magdev_frame.drop('good-pts', 1)
    magdev_frame = magdev_frame[magdev_frame.sd != 0.0]  # in practice this deletes the row containing COMP


    # Make a linear model of mag vs log(sd) using sklearn RANSAC

    log_sd = np.log10(magdev_frame[['sd']])
    rel_mag = magdev_frame[['mag']]

    # set the random no seed to maintain consistency between runs
    ransac = linear_model.RANSACRegressor(max_trials=100, random_state=99)
    ransac.fit(rel_mag, log_sd)

    # run a prediction
    pred_y = ransac.predict(rel_mag)

    residuals = log_sd['sd'].values - np.reshape(pred_y, (-1,))
    mad = median_absolute_deviation(residuals, ignore_nan=True)

    # rerun RANSAC, but this time with a factor for the deviation accepted
    # a factor of 5 has been applied to MAD. This is empirical.
    ransac = linear_model.RANSACRegressor(max_trials=200, random_state=99, residual_threshold=outlier_th*mad*5)
    ransac.fit(rel_mag, log_sd)

    # run a prediction
    pred_y = ransac.predict(rel_mag)


    # these are arrays of boolean that serve to identify which points are inliers and outliers
    inlier_mask = ransac.inlier_mask_
    outlier_mask = np.logical_not(inlier_mask)


    # add both actual and predicted log sd to the dataframe
    magdev_frame = magdev_frame.assign(log_sd=log_sd, pred_log_sd=np.reshape(pred_y, (-1,)))

    # plot magdev for all stars

    fig, ax1 = plt.subplots(figsize=(12,6))

    plt.rc('grid', linestyle='dotted', color='gray')
    ax1.grid(True)

    ax1.plot(rel_mag[inlier_mask], log_sd[inlier_mask], 'b.', label='inliers')
    ax1.plot(rel_mag[outlier_mask], log_sd[outlier_mask], 'r.', label='outliers')
    ax1.plot(rel_mag, pred_y, color='maroon', linewidth=1, label='best fit')
    ax1.set_xlabel('Mag relative to VAR')

    ax1.set_ylabel('log SD')
    ax1.grid(axis='x', linestyle='dotted', color='gray')


    plt.title('log S.D. vs Magnitude - all stars')
    plt.legend(loc='lower right')


    plt.text(0.02, 0.95, 'Target: ' + target_name, \
             horizontalalignment='left', transform=ax1.transAxes)
    plt.text(0.02, 0.9, 'Filter: ' + filter, \
             horizontalalignment='left', transform=ax1.transAxes)
    plt.text(0.02, 0.85, 'Obs Date: ' + init_time.iso, \
             horizontalalignment='left', transform=ax1.transAxes)
    plt.text(0.0, -0.1, observatory + ' Observatory', color='gray', \
             horizontalalignment='left', transform=ax1.transAxes)

    plt.savefig('all_stars_magdev.png', dpi=72)
    plt.close("all")  # clears memory

    ##############################################################
    # identify ids of the stars that are outliers

    outlier_ids = magdev_frame[['id']][outlier_mask]

    # Plot LC of the outlier stars closer than proximate dist  ************
    for outlier_star_id in outlier_ids['id'].values:
        if str(outlier_star_id) in proximate_stars:
            print('.', end='', flush=True)

            idx = np.where(all_stars == str(outlier_star_id))[0][0]

            cur_sel_df = df2_list[idx]  # dataset for currently selected star
            cur_sel_dfi = cur_sel_df.fillna(cur_sel_df.mean(), axis=0)  # condition to fill in NaN values

            vc_mag = cur_sel_df[['mag']] - comp_df[['mag']]     # V-C

            vc_err_sq = cur_sel_df[['err']]**2 + comp_df[['err']]**2
            vc_err = np.sqrt(vc_err_sq['err'])
            vc_df = pd.concat([cur_sel_df[['bjd']],vc_mag, vc_err], axis=1) # make new df with results

            bjd = vc_df['bjd'].values       # convert to numpy array
            mag = vc_df['mag'].values
            err = vc_df['err'].values
            meanmag = np.nanmean(mag)
            std = np.nanstd(mag, ddof=1)

            title = 'Outlier star id#' + str(outlier_star_id)
            filename = os.path.join(in_path, std_name+'_outlier-star_'+ str(outlier_star_id) +'.png')

            plot_lc_err(bjd, mag, err, meanmag, std, title, filename)




###########################################################################################
# Write data needed to add new timeseries to ExoFOP-TESS to the PESTphot_info.txt file
###########################################################################################

f_pcinfo.write('\n\n\n############################################################################\n')
f_pcinfo.write('Information below is for filling out the Add Timeseries form on ExoFOP-TESS\n')
f_pcinfo.write('############################################################################\n\n')



f_pcinfo.write('Telescope: ' + observatory + ' ' + telescope + '\n')
f_pcinfo.write('Camera: ' + camera + '\n')
f_pcinfo.write('Filter: ' + filter + '\n')
f_pcinfo.write('Pixel Scale (arcsec): ' + str(scale) + '\n')
f_pcinfo.write('PSF (arcsec): ' + str(round(fwhm_mean*scale,2)) + '\n')
f_pcinfo.write('Photometry aperture rad (pix): ' + str(phot_rad) + '\n')

try:  # the plate solve section and EB checking defines max_dmag
    f_pcinfo.write('dmag: ' + str(round(max_dmag, 1)) + '\n')
except:
    f_pcinfo.write('dmag: not calculated because plate solve switched off' + '\n')

f_pcinfo.write('Date: ' + obs_date + '\n')
f_pcinfo.write('Duration (m): ' + str(round(obs_dur)) + '\n')
f_pcinfo.write('# of Obs: ' + str(frm) + '\n')
f_pcinfo.write('Tag: ' + obs_date_strip + '_' + tess_user + '_' + observatory + '_1\n')


###########################################################################################





##################################################################################################
# Done!
f_pcinfo.close()
print('\nPESTphot completed sucessfully using date type: ' + date_type + ' and aperture no. '+ str(sel_ap)+'\n')

